from time import time

from .path import single_source_shortest_path
from .classes import Column
from .consts import EPSILON, MAX_LABEL_COST, MIN_COL_VOL


__all__ = ['find_ue', 'perform_column_generation', 'perform_network_assignment']


def _update_link_cost_array(spnetworks):
    """ update generalized link costs for each SPNetwork """
    for sp in spnetworks:
        tau = sp.get_demand_period().get_id()
        vot = sp.get_agent_type().get_vot()

        for link in sp.get_links():
            if not link.length:
                break

            sp.link_cost_array[link.get_seq_no()] = (
                link.get_generalized_cost(tau, vot)
            )


def _update_link_travel_time(links):
    for link in links:
        if not link.length:
            break

        link.calculate_td_vdf()


def _update_link_and_column_volume(column_pool, links, iter_num, reduce_path_vol = True):
    if not iter_num:
        return

    for link in links:
        if not link.length:
            break

        link.reset_period_flow_vol()

    for k, cv in column_pool.items():
        # k= (at, tau, oz_id, dz_id)
        tau = k[1]

        for col in cv.get_columns():
            path_vol = col.get_volume()
            for i in col.links:
                # 04/10/22, remove it from now on as it is not in need
                # pce_ratio = 1 and path_vol * pce_ratio
                links[i].increase_period_flow_vol(tau, path_vol)

            # 06/11/24: if we check cv.is_route_fixed() here,
            # 1. shall we check it anywhere else??
            # 2. is it really necessary?
            if reduce_path_vol and not cv.is_route_fixed():
                col.vol *= iter_num / (iter_num + 1)


def _update_column_gradient_cost_and_flow(column_pool, links, agent_types, iter_num):
    total_gap = 0
    total_sys_travel_time = 0

    for k, cv in column_pool.items():
        # k = (at, tau, oz_id, dz_id)
        vot = agent_types[k[0]].get_vot()
        tau = k[1]

        least_gradient_cost = MAX_LABEL_COST
        least_gradient_cost_path_id = -1

        for col in cv.get_columns():
            # i is link sequence no
            path_gradient_cost = sum(
                links[i].get_generalized_cost(tau, vot) for i in col.links
            )
            col.set_gradient_cost(path_gradient_cost)

            # col.get_volume() >= EPSILON is the key to eliminate ultra-low-volume
            # columns. along with the blow new_vol reset if new_vol < MIN_COL_VOL,
            # it enables shifting volume from the shortest path to a non-shortest path.
            if path_gradient_cost < least_gradient_cost and col.get_volume() >= EPSILON:
                least_gradient_cost = path_gradient_cost
                least_gradient_cost_path_id = col.get_id()

        total_switched_out_path_vol = 0
        if cv.get_column_num() >= 2:
            for col in cv.get_columns():
                if col.get_id() == least_gradient_cost_path_id:
                    continue

                if not col.get_volume():
                    continue

                col.update_gradient_cost_diffs(least_gradient_cost)

                # with the forgoing col.get_volume() >= EPSILON,
                # col.get_cap() could be negative
                total_gap += col.get_gap()
                total_sys_travel_time += col.get_sys_travel_time()

                step_size = 1 / (iter_num + 2) * cv.get_od_volume()
                cur_vol = col.get_volume()

                new_vol = cur_vol - step_size * col.get_gradient_cost_rel_diff()
                if new_vol < MIN_COL_VOL:
                    new_vol = 0

                col.set_volume(new_vol)
                total_switched_out_path_vol += cur_vol - new_vol

        if least_gradient_cost_path_id != -1:
            col = cv.get_column(least_gradient_cost_path_id)
            total_sys_travel_time += col.get_sys_travel_time()
            col.increase_volume(total_switched_out_path_vol)

    rel_gap = total_gap / max(total_sys_travel_time, EPSILON)

    print(f'current iteration number in column update: {iter_num}\n'
          f'total gap: {total_gap:.4e}; relative gap: {rel_gap:.4%}')


def _backtrace_shortest_path_tree(centroid,
                                  centroids,
                                  links,
                                  node_preds,
                                  link_preds,
                                  at_id,
                                  dp_id,
                                  column_pool,
                                  iter_num):

    if not centroid.has_outgoing_links():
        return

    oz_id = centroid.get_zone_id()
    k_path_prob = 1 / (iter_num + 1)

    for c in centroids:
        dz_id = c.get_zone_id()
        if dz_id == oz_id:
            continue

        if (at_id, dp_id, oz_id, dz_id) not in column_pool:
            continue

        cv = column_pool[(at_id, dp_id, oz_id, dz_id)]
        if cv.is_route_fixed():
            continue

        link_path = []
        dist = 0

        curr_node_no = c.get_node_no()
        # retrieve the sequence backwards
        while curr_node_no >= 0:
            curr_link_no = link_preds[curr_node_no]
            if curr_link_no >= 0:
                link_path.append(curr_link_no)
                dist += links[curr_link_no].length

            curr_node_no = node_preds[curr_node_no]

        # make sure this is a valid path
        if not link_path:
            continue

        vol = k_path_prob * cv.get_od_volume()

        existing = False
        for col in cv.get_columns():
            if col.get_distance() != dist:
                continue

            # the first and the last are connectors
            if col.get_links() == link_path[1:-1]:
                col.increase_volume(vol)
                existing = True
                break

        if not existing:
            path_id = cv.get_column_num()
            col = Column(path_id)
            col.set_volume(vol)
            col.set_distance(dist)
            col.links = [x for x in link_path[1:-1]]
            cv.add_new_column(col)


def _update_column_attributes(column_pool, links, agent_types):
    """ update toll and travel time for each column, and compute final UE convergency """
    total_gap = 0
    total_sys_travel_time = 0

    for k, cv in column_pool.items():
        # k = (at, dp, oz, dz)
        dp = k[1]
        vot = agent_types[k[0]].get_vot()
        least_gradient_cost = MAX_LABEL_COST

        i = 0
        for col in cv.get_columns():
            if not col.get_volume():
                col.reset()
                continue

            # reset column id
            col.set_id(i)
            i += 1

            nodes = []
            path_toll = 0
            travel_time = 0

            path_gradient_cost = 0
            for j in col.links:
                link = links[j]
                nodes.append(link.to_node_no)
                travel_time += link.travel_time_by_period[dp]
                path_toll += link.get_toll()
                path_gradient_cost += link.get_generalized_cost(dp, vot)

            # last node
            nodes.append(links[col.links[-1]].from_node_no)

            col.set_gradient_cost(path_gradient_cost)
            col.set_travel_time(travel_time)
            col.set_toll(path_toll)
            col.nodes = [x for x in nodes]

            if path_gradient_cost < least_gradient_cost:
                least_gradient_cost = path_gradient_cost

        for col in cv.get_columns():
            if not col.get_volume():
                continue

            col.update_gradient_cost_diffs(least_gradient_cost)
            total_sys_travel_time += col.get_sys_travel_time()
            total_gap += col.get_gap()

    rel_gap = total_gap / max(total_sys_travel_time, EPSILON)
    print('current iteration number in column update: postprocessing\n'
          f'total gap: {total_gap:.4e}; relative gap: {rel_gap:.4%}\n')


def _generate(spn, column_pool, iter_num):
    for c in spn.get_orig_centroids():
        node_id = c.get_node_id()
        single_source_shortest_path(spn, node_id)

        _backtrace_shortest_path_tree(c,
                                      spn.get_centroids(),
                                      spn.get_links(),
                                      spn.get_node_preds(),
                                      spn.get_link_preds(),
                                      spn.get_agent_type().get_id(),
                                      spn.get_demand_period().get_id(),
                                      column_pool,
                                      iter_num)


def _generate_column_pool(spnetworks, column_pool, iter_num):
    # single processing
    # it could be multiprocessing
    for spn in spnetworks:
        _generate(spn, column_pool, iter_num)


def perform_column_generation(column_gen_num, column_update_num, ui):
    """ perform network assignment using the selected assignment mode

    WARNING
    -------
    DEPRECATED due to its confusing name. Use find_ue() instead!

    Only Path/Column-based User Equilibrium (UE) is implemented in Python.
    If you need other assignment modes or dynamic traffic assignment (DTA),
    please use perform_network_assignment_DTALite()

    Parameters
    ----------
    column_gen_num
        number of iterations to be performed on generating column pool
        column pool

    column_update_num
        number of iterations to be performed on optimizing column pool

    ui
        network object generated by pg.read_network()

    Returns
    -------
    None

    Note
    ----
    You will need to call output_columns() and output_link_performance() to
    get the assignment results, i.e., paths/columns (in route_assignment.csv) and
    assigned volumes and other link attributes on each link (in link_performance.csv)
    """
    # make sure iteration numbers are both non-negative
    assert(column_gen_num>=0)
    assert(column_update_num>=0)

    # base assignment
    A = ui._base_assignment
    # set up SPNetwork
    A.setup_spnetwork()

    links = A.get_links()
    ats = A.get_agent_types()
    column_pool = A.get_column_pool()

    print('find user equilibrium (UE)')
    st = time()

    for i in range(column_gen_num):
        print(f'current iteration number in column generation: {i}')

        _update_link_and_column_volume(column_pool, links, i)
        _update_link_travel_time(links)
        # update generalized link cost before assignment
        _update_link_cost_array(A.get_spnetworks())
        # loop through all centroids on the base network
        _generate_column_pool(A.get_spnetworks(), column_pool, i)

    print(f'\nprocessing time of generating columns: {time()-st:.2f} s\n')

    for i in range(column_update_num):
        _update_link_and_column_volume(column_pool, links, i, False)
        _update_link_travel_time(links)
        _update_column_gradient_cost_and_flow(column_pool, links, ats, i)

    # postprocessing
    _update_link_and_column_volume(column_pool, links, column_gen_num, False)
    _update_link_travel_time(links)
    _update_column_attributes(column_pool, links, ats)


def update_links_using_columns(network):
    """ a helper function for load_columns() """
    A = network._base_assignment

    column_pool = A.get_column_pool()
    links = A.get_links()

    # do not update column volume
    _update_link_and_column_volume(column_pool, links, 1, False)
    _update_link_travel_time(links)


def perform_network_assignment(assignment_mode, column_gen_num, column_update_num, ui):
    """DEPRECATED Column Generation API. Use find_ue() instead!

    Keep it here as legacy support for existing users who already get used to it
    """
    print('This function has been deprecated, and will be removed later!'
          'Please use perform_column_generation() instead.')

    if assignment_mode != 1:
        raise Exception(
            'NOT implemented yet!'
            'Please please use perform_network_assignment_DTALite().'
        )

    perform_column_generation(column_gen_num, column_update_num, ui)


def find_ue(ui, column_gen_num, column_update_num):
    """ find user equilibrium (UE)

    WARNING
    -------
    Only Path/Column-based User Equilibrium (UE) is implemented in Python.
    If you need other assignment modes or dynamic traffic assignment (DTA),
    please use perform_network_assignment_DTALite()

    Parameters
    ----------
    ui
        network object generated by pg.read_network()

    column_gen_num
        number of iterations to be performed on generating column pool
        column pool

    column_update_num
        number of iterations to be performed on optimizing column pool

    Returns
    -------
    None

    Note
    ----
    You will need to call output_columns() and output_link_performance() to
    get the assignment results, i.e., paths/columns (in route_assignment.csv) and
    volumes and other link attributes on each link (in link_performance.csv)
    """
    perform_column_generation(column_gen_num, column_update_num, ui)


from os import chdir, getcwd

from path4gmns.dtaapi import perform_network_assignment_DTALite, run_DTALite


ORIG_DIR = getcwd()


def test_classic_dtalite(sample_data_dir, mode = 1):
    chdir(sample_data_dir)

    column_gen_num = 20
    column_update_num = 20
    perform_network_assignment_DTALite(mode, column_gen_num, column_update_num)

    chdir(ORIG_DIR)


# disable it from now on (10/16/24) as run_DTALite() requires a different
# settings.yml
# def test_multimodal_dtalite(sample_data_dir):
#     chdir(sample_data_dir)

#     run_DTALite()

#     chdir(ORIG_DIR)


import ctypes
from collections import deque
from copy import deepcopy
from datetime import datetime
from math import ceil, floor
from random import choice, randint

from .consts import EPSILON, MAX_LABEL_COST, SECONDS_IN_MINUTE, SECONDS_IN_HOUR
from .path import find_path_for_agents, find_shortest_path, \
                  single_source_shortest_path, benchmark_apsp


__all__ = ['UI']


class Node:

    def __init__(self, node_no, node_id, zone_id, x='', y='', is_activity_node=False):
        """ the attributes of node  """
        # node_no: internal node index used for calculation
        self.node_no = node_no
        # node_id: user defined node id from input
        self.node_id = node_id
        # link objects
        self.outgoing_links = []
        self.incoming_links = []
        self.zone_id = zone_id
        self.coord_x = x
        self.coord_y = y
        self.is_activity_node = is_activity_node

    def has_outgoing_links(self):
        return len(self.outgoing_links) > 0

    def has_incoming_links(self):
        return len(self.incoming_links) > 0

    def get_zone_id(self):
        return self.zone_id

    def get_node_id(self):
        return self.node_id

    def get_node_no(self):
        return self.node_no

    def get_coordinate(self):
        return self.coord_x + ' ' + self.coord_y

    def add_outgoing_link(self, link):
        self.outgoing_links.append(link)

    def add_incoming_link(self, link):
        self.incoming_links.append(link)

    def update_coordinate(self, args):
        self.coord_x = args[0]
        self.coord_y = args[1]

    def get_incoming_link_num(self):
        return len(self.incoming_links)


class Link:

    def __init__(self,
                 id,
                 link_no,
                 from_node_no,
                 to_node_no,
                 from_node_id,
                 to_node_id,
                 length,
                 lanes=1,
                 link_type=1,
                 free_speed=60,
                 capacity=1999,
                 toll = 0,
                 allowed_uses='all',
                 geometry='',
                 demand_period_size=1):
        """ the attributes of link """
        self.id = id
        self.link_no = link_no
        self.from_node_no = from_node_no
        self.to_node_no = to_node_no
        self.from_node_id = from_node_id
        self.to_node_id = to_node_id
        # length is mile or km
        self.length = length
        self.lanes = lanes
        # 1: one direction, 2: two way
        self.type = link_type
        # free flow travel time in minutes, free_speed is either mile/h or km/h
        self.fftt = (
            length / max(EPSILON, free_speed) * 60
        )
        # capacity is lane capacity per hour
        self.link_capacity = capacity * lanes
        self.allowed_uses = allowed_uses
        self.geometry = geometry
        # add for CG
        self.demand_period_size = demand_period_size
        self.toll = toll
        self.route_choice_cost = 0
        self.travel_time_by_period = [0] * demand_period_size
        self.flow_vol_by_period = [0] * demand_period_size
        self.vdfperiods = []
        # for simulation
        self.cum_arr = None
        self.cum_dep = None
        self.outflow_cap = None
        self.waiting_time = None
        self.entr_queue = deque()
        self.exit_queue = deque()
        # for ODME
        self.obs = 0
        self.est_dev = 0
        self.is_obs_upper_bounded = False

    def get_link_id(self):
        return self.id

    def get_seq_no(self):
        return self.link_no

    def get_from_node_id(self):
        return self.from_node_id

    def get_to_node_id(self):
        return self.to_node_id

    def get_length(self):
        return self.length

    def get_geometry(self):
        return self.geometry

    def get_toll(self):
        return self.toll

    def get_free_flow_travel_time(self):
        return self.fftt

    def get_route_choice_cost(self):
        return self.route_choice_cost

    def get_period_travel_time(self, tau):
        return self.travel_time_by_period[tau]

    def get_period_flow_vol(self, tau):
        return self.flow_vol_by_period[tau]

    def get_period_voc(self, tau):
        return self.vdfperiods[tau].get_voc()

    def get_period_fftt(self, tau):
        try:
            return self.vdfperiods[tau].get_fftt()
        except IndexError:
            raise Exception(f'NO such demand period id: {tau}!'
                            ' Check your input demand_period_id')

    def get_period_avg_travel_time(self, tau):
        return self.vdfperiods[tau].get_avg_travel_time()

    def get_generalized_cost(self, tau, value_of_time):
        return (
            self.travel_time_by_period[tau]
            + self.route_choice_cost
            + self.toll / max(EPSILON, value_of_time) * 60
        )

    def reset_period_flow_vol(self):
        for tau in range(self.demand_period_size):
            self.flow_vol_by_period[tau] = 0

    def increase_period_flow_vol(self, tau, fv):
        self.flow_vol_by_period[tau] += fv

    def calculate_td_vdf(self):
        for tau in range(self.demand_period_size):
            self.travel_time_by_period[tau] = (
                self.vdfperiods[tau].run_bpr(self.flow_vol_by_period[tau])
            )

    def update_waiting_time(self, minute, wt):
        try:
            self.waiting_time[minute] += wt
        except IndexError:
            pass

    def set_capacity_ratio(self, tau, rr):
        self.vdfperiods[tau].capacity_ratio = rr


class Agent:
    """ individual agent derived from aggregated demand between an OD pair

    id: integer starts from 1
    seq_no: internal agent index starting from 0 used for calculation
    """
    def __init__(self, agent_id, agent_no, agent_type_id, demand_period_id,
                 o_zone_id, d_zone_id):
        """ the attribute of agent """
        self.id = agent_id
        self.seq_no = agent_no
        # vehicle
        self.at_id = agent_type_id
        self.dp_id = demand_period_id
        self.o_zone_id = o_zone_id
        self.d_zone_id = d_zone_id
        self.o_node_id = 0
        self.d_node_id = 0
        self.node_path = None
        self.link_path = None
        # Passenger Car Equivalent (PCE) of the agent
        self.PCE_factor = 1
        self.path_cost = 0
        # for simulation
        self.dep_time = 0
        # corresponding to each link along the link path
        self.link_dep_interval = None
        self.link_arr_interval = None
        self.curr_link_pos = 0

    def get_orig_node_id(self):
        return self.o_node_id

    def get_dest_node_id(self):
        return self.d_node_id

    def get_seq_no(self):
        return self.seq_no

    def get_id(self):
        return self.id

    def get_orig_zone_id(self):
        return self.o_zone_id

    def get_dest_zone_id(self):
        return self.d_zone_id

    def get_path_cost(self):
        return self.path_cost

    def get_node_path(self):
        return self.node_path

    def get_at_id(self):
        return self.at_id

    def get_dp_id(self):
        return self.dp_id

    def get_od(self):
        return self.o_zone_id, self.d_zone_id

    def update_dep_interval(self, intvl):
        self.link_dep_interval[self.curr_link_pos] = (
            self.link_arr_interval[self.curr_link_pos] + intvl
        )

    def get_curr_dep_interval(self):
        return self.link_dep_interval[self.curr_link_pos]

    def reached_last_link(self):
        return self.curr_link_pos == 0

    def get_next_link_no(self):
        pos = self.curr_link_pos - 1
        return self.link_path[pos]

    def set_dep_interval(self, i):
        self.link_dep_interval[self.curr_link_pos] = i

    def set_arr_interval(self, i, increment=0):
        self.link_arr_interval[self.curr_link_pos-increment] = i

    def get_arr_interval(self):
        return self.link_arr_interval[self.curr_link_pos]

    def increment_link_pos(self):
        if self.curr_link_pos > 0:
            self.curr_link_pos -= 1

    def get_dep_time(self):
        return self.dep_time

    def get_origin_dep_interval(self):
        return self.link_arr_interval[-1]


class Zone:

    def __init__(self, zone_id, bin_index=0):
        self.no = -1
        self.id = str(zone_id)
        self.bin_id = bin_index
        self.production = 0
        self.centroid = None
        self.boundaries = []
        self.coord_x = 91
        self.coord_y = 181
        self.nodes = []
        self.activity_nodes = []
        # for ODME
        self.prod_obs = 0
        self.attr_obs = 0
        self.attr_est = 0
        self.prod_est = 0
        self.prod_est_dev = 0
        self.attr_est_dev = 0
        self.is_prod_obs_upper_bounded = False
        self.is_attr_obs_upper_bounded = False

    def get_activity_nodes(self):
        return self.activity_nodes

    def get_activity_nodes_num(self):
        return len(self.activity_nodes)

    def get_bin_index(self):
        return self.bin_id

    def get_boundaries(self):
        return self.boundaries

    def get_centroid(self):
        return self.centroid

    def get_coordinate(self):
        return self.coord_x, self.coord_y

    def get_coordinate_str(self):
        if self.coord_x == 91 or self.coord_y == 181:
            return ''

        return str(self.coord_x) + ' ' + str(self.coord_y)

    def get_geo(self):
        """ output the four vertices as boundary """
        try:
            [U, D, L, R] = self.get_boundaries()
            geo = (
                'LINESTRING ('
                + str(L) + ' ' + str(U) + ','
                + str(R) + ' ' + str(U) + ','
                + str(R) + ' ' + str(D) + ','
                + str(L) + ' ' + str(D) + ','
                + str(L) + ' ' + str(U) + ')'
            )
        except ValueError:
            geo = 'LINESTRING ()'

        return geo

    def get_nodes(self):
        return self.nodes

    def get_production(self):
        return self.production

    def add_activity_node(self, node_id):
        self.activity_nodes.append(node_id)

    def add_node(self, node_id):
        self.nodes.append(node_id)

    def set_bin_index(self, bi):
        self.bin_id = bi

    def set_coord(self, cx, cy):
        self.coord_x = cx
        self.coord_y = cy

    def set_geo(self, U, D, L, R):
        self.boundaries = [U, D, L, R]

    def set_production(self, p):
        self.production = p


class Network:

    def __init__(self):
        self.nodes = []
        self.links = []
        self.agents = []
        self.node_size = 0
        self.link_size = 0
        self.agent_size = 0
        # key: node id, value: node seq no
        self.map_id_to_no = {}
        # key: node seq no, value: node id
        self.map_no_to_id = {}
        # map link id to link seq no
        self.link_ids = {}
        self.node_label_cost = None
        self.node_preds = None
        self.link_preds = None
        self.capi_allocated = False
        self.agent_type_name = 'all'
        # key: zone id, value: zone object
        self.zones = {}
        self.activity_node_num = 0
        self.last_thru_node = 0
        self.centroids_added = False
        # key: simulation interval, value: agent id
        self.td_agents = {}

    def update(self):
        self.node_size = len(self.nodes)
        self.link_size = len(self.links)
        self.agent_size = len(self.agents)
        self.last_thru_node = self.node_size
        self.activity_node_num = sum(
            z.get_activity_nodes_num() for z in self.zones.values()
        )

    def allocate_for_CAPI(self):
        # execute only on the first call
        if self.capi_allocated:
            return

        node_size = self.node_size
        link_size = self.link_size

        # initialization for predecessors and label costs
        node_preds = [-1] * node_size
        link_preds = [-1] * node_size
        node_label_cost = [MAX_LABEL_COST] * node_size

        # initialize from_node_no_array, to_node_no_array, and link_cost_array
        from_node_no_array = [link.from_node_no for link in self.links]
        to_node_no_array = [link.to_node_no for link in self.links]
        link_cost_array = [link.fftt for link in self.links]

        # initialize others
        queue_next = [0] * node_size
        first_link_from = [-1] * node_size
        last_link_from = [-1] * node_size
        sorted_link_no_array = [-1] * link_size

        # internal link index used for shortest path calculation only
        j = 0
        for i, node in enumerate(self.nodes):
            if not node.outgoing_links:
                continue
            first_link_from[i] = j
            for link in node.outgoing_links:
                # set up the mapping from j to the true link seq no
                sorted_link_no_array[j] = link.link_no
                j += 1
            last_link_from[i] = j

        # setup allowed uses
        allowed_uses = [link.allowed_uses for link in self.links]

        # set up arrays using ctypes
        int_arr_node = ctypes.c_int * node_size
        int_arr_link = ctypes.c_int * link_size
        double_arr_node = ctypes.c_double * node_size
        double_arr_link = ctypes.c_double * link_size
        # for allowed_uses
        char_arr_link = ctypes.c_wchar_p * link_size

        self.from_node_no_array = int_arr_link(*from_node_no_array)
        self.to_node_no_array = int_arr_link(*to_node_no_array)
        self.first_link_from = int_arr_node(*first_link_from)
        self.last_link_from = int_arr_node(*last_link_from)
        self.sorted_link_no_array = int_arr_link(*sorted_link_no_array)
        self.link_cost_array = double_arr_link(*link_cost_array)
        self.node_label_cost = double_arr_node(*node_label_cost)
        self.node_preds = int_arr_node(*node_preds)
        self.link_preds = int_arr_node(*link_preds)
        self.queue_next = int_arr_node(*queue_next)
        self.allowed_uses = char_arr_link(*allowed_uses)

        self.capi_allocated = True

    def add_centroids_connectors(self):
        if self.centroids_added:
            return

        node_no = self.node_size
        link_no = self.link_size
        # get zones
        for z in self.get_zones():
            if not z:
                continue

            # create a centroid
            node_id = 'c_' + z
            centroid = Node(node_no, node_id, z)
            # try coordinate of the centroid from each zone first. if the values
            # are invalid, then use that of the first node from each zone.
            coord_x, coord_y = self.zones[z].get_coordinate()
            if coord_x == 91 or coord_y == 181:
                node_id_ = self.get_nodes_from_zone(z)[0]
                node_no_ = self.get_node_no(node_id_)
                node = self.get_nodes()[node_no_]
                coord_x = node.coord_x
                coord_y = node.coord_y
                self.zones[z].set_coord(float(coord_x), float(coord_y))
            else:
                coord_x = str(coord_x)
                coord_y = str(coord_y)

            centroid.update_coordinate((coord_x, coord_y))
            self.zones[z].centroid = centroid

            self.nodes.append(centroid)
            self.map_id_to_no[node_id] = node_no
            self.map_no_to_id[node_no] = node_id

            # build connectors
            for i in self.get_nodes_from_zone(z):
                link_id_f = 'conn_' + str(link_no)
                from_node_no = node_no
                to_node_id = i

                try:
                    to_node_no = self.map_id_to_no[i]
                except KeyError:
                    continue

                # connector from centroid to activity nodes in this zone
                c_forward = Link(link_id_f,
                                 link_no,
                                 from_node_no,
                                 to_node_no,
                                 node_id,
                                 to_node_id,
                                 0)

                # connector from activity nodes in this zone to centroid
                link_id_b = 'conn_' + str(link_no+1)
                c_backward = Link(link_id_b,
                                  link_no+1,
                                  to_node_no,
                                  from_node_no,
                                  to_node_id,
                                  node_id,
                                  0)

                self.nodes[from_node_no].add_outgoing_link(c_forward)
                self.nodes[to_node_no].add_outgoing_link(c_backward)

                self.links.append(c_forward)
                self.links.append(c_backward)

                link_no += 2

            node_no += 1

        self.node_size = len(self.nodes)
        self.link_size = len(self.links)
        self.centroids_added = True

    def setup_agents(self, column_pool):
        agent_id = 1
        agent_no = 0

        for k, cv in column_pool.items():
            if cv.get_od_volume() <= 0:
                continue

            # k= (at, dp, orig, dest)
            at = k[0]
            dp = k[1]
            oz = k[2]
            dz = k[3]

            vol = int(cv.get_od_volume()+1)
            for _ in range(vol):
                # construct agent using valid record
                agent = Agent(agent_id,
                              agent_no,
                              at,
                              dp,
                              oz,
                              dz)

                # step 1 generate o_node_id and d_node_id randomly
                # according to o_zone_id and d_zone_id
                agent.o_node_id = choice(
                    self.zones[oz].get_nodes()
                )
                agent.d_node_id = choice(
                    self.zones[dz].get_nodes()
                )

                # step 2 update agent_id and agent_seq_no
                agent_id += 1
                agent_no += 1

                self.agents.append(agent)

        self.agent_size = len(self.agents)
        print(f'the number of agents is {self.agent_size}')

    def _get_agent(self, agent_no):
        """ retrieve agent using agent_no """
        try:
            return self.agents[agent_no]
        except IndexError:
            agent_id = agent_no + 1
            print(f'Please provide a valid agent id. agent_id: {agent_id} does NOT EXIST!')

    def get_agent_node_path(self, agent_id, path_only):
        """ return the sequence of node IDs along the agent path

        developer's note: consider changing its name to
        get_agent_node_path_str()
        """
        agent_no = agent_id - 1
        agent = self._get_agent(agent_no)

        path_cost = agent.get_path_cost()
        if path_cost >= MAX_LABEL_COST:
            return f'distance: infinity | path: '

        path = ''
        if agent.node_path:
            path = ';'.join(
                self.map_no_to_id[x] for x in reversed(agent.node_path)
            )

        if path_only:
            return path
        else:
            return f'distance: {path_cost:.2f} mi | node path: {path}'

    def get_agent_link_path(self, agent_id, path_only):
        """ return the sequence of link IDs along the agent path

        developer's note: consider changing its name to
        get_agent_link_path_str()
        """
        agent_no = agent_id - 1
        agent = self._get_agent(agent_no)

        path_cost = agent.get_path_cost()
        if path_cost >= MAX_LABEL_COST:
            return f'distance: infinity | path: '

        path = ''
        if agent.link_path:
            path = ';'.join(
                self.links[x].get_link_id() for x in reversed(agent.link_path)
            )

        if path_only:
            return path
        else:
            return f'distance: {path_cost:.2f} mi | link path: {path}'

    def get_agent_orig_node_id(self, agent_id):
        """ return the origin node id of agent """
        agent_no = agent_id - 1
        agent = self._get_agent(agent_no)

        return agent.get_orig_node_id()

    def get_agent_dest_node_id(self, agent_id):
        """ return the origin node id of agent """
        agent_no = agent_id - 1
        agent = self._get_agent(agent_no)

        return agent.get_dest_node_id()

    def get_agent_count(self):
        return self.agent_size

    def get_nodes_from_zone(self, zone_id):
        return self.zones[zone_id].get_nodes()

    def get_node_no(self, node_id):
        return self.map_id_to_no[node_id]

    def get_node_size(self):
        return self.node_size

    def get_link_size(self):
        return self.link_size

    def get_nodes(self):
        return self.nodes

    def get_links(self):
        return self.links

    def get_zones(self):
        # sorting is needed for setup_spnetwork()
        return sorted(self.zones.keys())

    def get_zone_size(self):
        return len(self.zone_)

    def get_from_node_no_arr(self):
        return self.from_node_no_array

    def get_to_node_no_arr(self):
        return self.to_node_no_array

    def get_first_links(self):
        return self.first_link_from

    def get_last_links(self):
        return self.last_link_from

    def get_sorted_link_no_arr(self):
        return self.sorted_link_no_array

    def get_node_preds(self):
        return self.node_preds

    def get_link_preds(self):
        return self.link_preds

    def get_node_label_costs(self):
        return self.node_label_cost

    def get_link_costs(self):
        return self.link_cost_array

    def get_queue_next(self):
        return self.queue_next

    def get_allowed_uses(self):
        return self.allowed_uses

    def get_link(self, seq_no):
        return self.links[seq_no]

    def get_agent_type_name(self):
        """ for allowed uses in single_source_shortest_path()"""
        return self.agent_type_name

    def get_link_no(self, id):
        return self.link_ids[id]

    def get_agents(self):
        return self.agents

    def get_last_thru_node(self):
        """ node no of the first potential centroid """
        return self.last_thru_node

    def set_agent_type_name(self, at_name):
        self.agent_type_name = at_name

    def get_centroids(self):
        for k, v in self.zones.items():
            if not k:
                continue

            yield v.get_centroid()

    def have_dep_agents(self, i):
        return i in self.td_agents

    def get_td_agents(self, i):
        return self.td_agents[i]


class Column:
    """ column is path """
    def __init__(self, id=-1):
        self.id = id
        self.vol = 0
        self.dist = 0
        self.toll = 0
        self.travel_time = 0
        self.gradient_cost = 0
        self.gradient_cost_abs_diff = 0
        self.gradient_cost_rel_diff = 0
        self.nodes = None
        self.links = None
        self.geo = ''

    def get_link_num(self):
        return len(self.links)

    def get_node_num(self):
        return len(self.nodes)

    def get_id(self):
        return self.id

    def get_distance(self):
        return self.dist

    def get_volume(self):
        return self.vol

    def get_toll(self):
        return self.toll

    def get_travel_time(self):
        return self.travel_time

    def get_gradient_cost(self):
        return self.gradient_cost

    def get_gradient_cost_abs_diff(self):
        return self.gradient_cost_abs_diff

    def get_gradient_cost_rel_diff(self):
        return self.gradient_cost_rel_diff

    def get_links(self):
        """ return link path """
        return self.links

    def get_nodes(self):
        """ return node path """
        return self.nodes

    def set_distance(self, d):
        self.dist = d

    def set_volume(self, v):
        self.vol = v

    def set_toll(self, t):
        self.toll = t

    def set_travel_time(self, tt):
        self.travel_time = tt

    def set_gradient_cost(self, c):
        self.gradient_cost = c

    def increase_volume(self, v):
        self.vol += v

    def set_geometry(self, g):
        self.geo = g

    def update_gradient_cost_diffs(self, least_gc):
        self.gradient_cost_abs_diff = self.gradient_cost - least_gc
        self.gradient_cost_rel_diff = (
            self.gradient_cost_abs_diff / max(EPSILON, least_gc)
        )

    def get_gap(self):
        return self.gradient_cost_abs_diff * self.vol

    def get_sys_travel_time(self):
        return self.gradient_cost * self.vol

    def reset(self):
        self.id = -1
        self.nodes = self.links = None
        self.travel_time = self.gradient_cost = 0
        self.gradient_cost_abs_diff = self.gradient_cost_rel_diff = 0

    def set_id(self, id):
        self.id = id


class ColumnVec:
    """ column pool for (at, dp, oz, dz)

    where, at is agent type id,
           dp is demand period id,
           oz is origin zone id,
           dz is destination zone id.
    """
    def __init__(self):
        self.od_vol = 0
        self.route_fixed = False
        self.node_seq_paths = []

    def is_route_fixed(self):
        return self.route_fixed

    def get_od_volume(self):
        return self.od_vol

    def get_column_num(self):
        return len(self.node_seq_paths)

    def get_columns(self):
        return self.node_seq_paths

    def get_column(self, k):
        return self.node_seq_paths[k]

    def add_new_column(self, col):
        self.node_seq_paths.append(col)

    def set_volume(self, vol):
        self.od_vol = vol

    def increase_volume(self, vol):
        self.od_vol += vol


class AgentType:

    def __init__(self, id=0, type='a', name='auto',
                 vot=10, flow_type=0, pce=1, ffs=60, use_link_ffs=True):
        """ default constructor """
        self.id = id
        self.type = type
        self.name = name
        self.vot = vot
        self.flow_type = flow_type
        self.pce = pce
        self.ffs = ffs
        self.use_link_ffs = use_link_ffs

    def get_id(self):
        return self.id

    def get_name(self):
        return self.name

    def get_vot(self):
        return self.vot

    def get_type_str(self):
        return self.type

    def get_pce(self):
        return self.pce

    def get_free_flow_speed(self):
        return self.ffs

    @staticmethod
    def get_default_type_str():
        return 'a'

    @staticmethod
    def get_default_name():
        return 'auto'

    @staticmethod
    def get_legacy_type_str():
        return 'p'

    @staticmethod
    def get_legacy_name():
        return 'passenger'


class SpecialEvent:

    def __init__(self, name) -> None:
        self.name = name
        self.affected_links = {}

    def get_affected_links(self):
        return self.affected_links.items()


class DemandPeriod:

    def __init__(self, id=0, period='AM', time_period='0700-0800'):
        self.id = id
        self.period = period
        self.time_period = time_period
        self.special_event = None
        self._setup_time()

    def _parse_time_period(self, delim='-'):
        s1, s2 = self.time_period.split(delim)
        t1 = datetime.strptime(s1, '%H%M')
        t2 = datetime.strptime(s2, '%H%M')
        if t2 <= t1:
            raise ValueError('ending time <= starting time')

        st_ = t1.hour * 60 + t1.minute
        et_ = t2.hour * 60 + t2.minute

        return st_, et_

    def _setup_time(self):
        try:
            st_, et_ = self._parse_time_period()
        except ValueError:
            # backward compatibility for versions <= v0.9.3
            st_, et_ = self._parse_time_period('_')
        except Exception as e:
            raise e

        self.st = st_
        self.et = et_

    def get_id(self):
        return self.id

    def get_period(self):
        return self.period

    def get_duration(self):
        """ duration of demand period in minutes """
        return self.et - self.st

    def get_start_time(self):
        return self.st


class Demand:

    def __init__(self, id=0, period='AM', agent_type='a', file='demand.csv'):
        self.id = id
        self.period = period
        self.agent_type_str = agent_type
        self.file = file

    def get_id(self):
        return self.id

    def get_file_name(self):
        return self.file

    def get_period(self):
        return self.period

    def get_agent_type_str(self):
        return self.agent_type_str


class VDFPeriod:

    def __init__(self, id, alpha=0.15, beta=4, mu=1000,
                 fftt=0, cap=99999, phf=-1):
        """ default constructor """
        self.id = id
        # the following four have been defined in class Link
        # they should be exactly the same with those in the corresponding link
        self.alpha = alpha
        self.beta = beta
        self.mu = mu
        # free flow travel time
        self.fftt = fftt
        self.capacity = cap
        self.phf = phf
        self.avg_travel_time = 0
        self.voc = 0
        self.capacity_ratio = 1

    def get_avg_travel_time(self):
        return self.avg_travel_time

    def get_voc(self):
        return self.voc

    def get_fftt(self):
        return self.fftt

    def run_bpr(self, vol):
        vol = max(0, vol)
        self.voc = vol / max(EPSILON, self.capacity * self.capacity_ratio)
        self.avg_travel_time = (
            self.fftt
            + self.fftt
            * self.alpha
            * pow(self.voc, self.beta)
        )

        return self.avg_travel_time


class SPNetwork(Network):
    """ attributes related to outputs from shortest path calculations """
    def __init__(self, base, at, dp):
        self.base = base
        self.nodes = self.base.get_nodes()
        self.links = self.base.get_links()
        self.node_size = base.get_node_size()
        self.link_size = base.get_link_size()
        # AgentType object
        self.agent_type = at
        # DemandPeriod object
        self.demand_period = dp
        # zone sequence no
        self.orig_zones = []
        self.capi_allocated = False
        super().allocate_for_CAPI()

    def allocate_for_CAPI(self):
        pass

    def get_node_no(self, node_id):
        return self.base.get_node_no(node_id)

    def get_agent_type(self):
        return self.agent_type

    def get_agent_type_name(self):
        # convert it to C char
        return self.agent_type.get_name()

    def get_demand_period(self):
        return self.demand_period

    def get_node_size(self):
        return super().get_node_size()

    def get_link_size(self):
        return super().get_link_size()

    def get_nodes(self):
        return super().get_nodes()

    def get_links(self):
        return super().get_links()

    def get_zones(self):
        return super().get_zones()

    def get_from_node_no_arr(self):
        return super().get_from_node_no_arr()

    def get_to_node_no_arr(self):
        return super().get_to_node_no_arr()

    def get_first_links(self):
        return super().get_first_links()

    def get_last_links(self):
        return super().get_last_links()

    def get_sorted_link_no_arr(self):
        return super().get_sorted_link_no_arr()

    def get_allowed_uses(self):
        return super().get_allowed_uses()

    def get_node_preds(self):
        return super().get_node_preds()

    def get_link_preds(self):
        return super().get_link_preds()

    def get_node_label_costs(self):
        return super().get_node_label_costs()

    def get_link_costs(self):
        return super().get_link_costs()

    def get_queue_next(self):
        return super().get_queue_next()

    # the following three are shared by all SPNetworks as the underlying 
    # network topology
    def get_last_thru_node(self):
        """ node no of the first potential centroid """
        return self.base.get_last_thru_node()

    def get_orig_centroids(self):
        for z in self.orig_zones:
            yield self.base.zones[z].get_centroid()

    def get_centroids(self):
        return self.base.get_centroids()


class AccessNetwork(Network):
    """ network for accessibility evaluation """
    def __init__(self, base, add_cc=True):
        self.base = base
        self.nodes = self.base.get_nodes()
        self.links = self.base.get_links()
        self.zones = self.base.zones
        self.map_id_to_no = self.base.map_id_to_no
        self.map_no_to_id = self.base.map_no_to_id
        self.node_size = self.base.get_node_size()
        self.link_size = self.base.get_link_size()
        self.centroids_added = self.base.centroids_added
        self.agent_type_name = 'all'
        self.pre_source_node_id = ''
        if add_cc:
            self._add_centroids_connectors()
        self.capi_allocated = False
        super().allocate_for_CAPI()

    def _add_centroids_connectors(self):
        if self.centroids_added:
            return

        # deep copy
        self.nodes = deepcopy(self.nodes)
        self.links = deepcopy(self.links)
        self.map_id_to_no = deepcopy(self.map_id_to_no)
        self.map_no_to_id = deepcopy(self.map_no_to_id)

        super().add_centroids_connectors()

    def get_zones(self):
        return self.base.get_zones()

    def get_nodes_from_zone(self, zone_id):
        return self.base.get_nodes_from_zone(zone_id)

    def _get_zone_coord(self, zone_id):
        """ coordinate of each zone is from its first node """
        node_id = self.get_nodes_from_zone(zone_id)[0]
        node_no = self.base.get_node_no(node_id)
        node = self.base.get_nodes()[node_no]
        return node.coord_x, node.coord_y

    def set_target_mode(self, mode):
        """ set up the target mode for accessibility evaluation

        Parameters
        ----------
        mode : agent name which is in settings.yml.
        """
        self.agent_type_name = mode

    def set_source_node_id(self, node_id):
        self.pre_source_node_id = node_id

    def get_agent_type_name(self):
        return self.agent_type_name

    def get_centroids(self):
        return self.base.get_centroids()

    def get_node_no(self, node_id):
        return self.map_id_to_no[node_id]

    def get_node_size(self):
        return super().get_node_size()

    def get_link_size(self):
        return super().get_link_size()

    def get_from_node_no_arr(self):
        return super().get_from_node_no_arr()

    def get_to_node_no_arr(self):
        return super().get_to_node_no_arr()

    def get_first_links(self):
        return super().get_first_links()

    def get_last_links(self):
        return super().get_last_links()

    def get_sorted_link_no_arr(self):
        return super().get_sorted_link_no_arr()

    def get_node_preds(self):
        return super().get_node_preds()

    def get_link_preds(self):
        return super().get_link_preds()

    def get_node_label_costs(self):
        return super().get_node_label_costs()

    # this function does not exist in class Network
    def get_node_label_cost(self, node_no):
        return self.node_label_cost[node_no]

    def get_link_costs(self):
        return super().get_link_costs()

    def get_queue_next(self):
        return super().get_queue_next()

    def get_allowed_uses(self):
        return super().get_allowed_uses()

    def get_last_thru_node(self):
        """ node no of the first centroid """
        return self.base.get_node_size()

    def get_pred_link_id(self, node_id):
        """ return id of the predecessor link to node_id """
        link_no = self.link_preds[self.get_node_no(node_id)]
        return self.links[link_no].get_link_id()

    def get_sp_distance(self, node_no):
        """ get the shortest path distance """
        if self.link_preds[node_no] == -1:
            return MAX_LABEL_COST

        dist = 0
        while node_no >= 0:
            link_no = self.link_preds[node_no]
            if link_no >= 0:
                dist += self.get_link(link_no).get_length()

            node_no = self.node_preds[node_no]

        return dist

    def update_generalized_link_cost(self, at, time_dependent, demand_period_id):
        """ update generalized link costs to calculate accessibility """
        vot = at.get_vot()

        if time_dependent:
            for link in self.get_links():
                # do not update connectors
                if link.get_link_id().startswith('conn_'):
                    continue

                self.link_cost_array[link.get_seq_no()] = (
                    link.get_period_fftt(demand_period_id)
                    + link.get_route_choice_cost()
                    + link.get_toll() / max(EPSILON, vot) * 60
                )
        else:
            if not at.use_link_ffs:
                ffs = at.get_free_flow_speed()

                for link in self.get_links():
                    self.link_cost_array[link.get_seq_no()] = (
                        (link.get_length() / max(EPSILON, ffs) * 60)
                        + link.get_route_choice_cost()
                        + link.get_toll() / max(EPSILON, vot) * 60
                    )
            else:
                for link in self.get_links():
                    self.link_cost_array[link.get_seq_no()] = (
                        link.get_free_flow_travel_time()
                        + link.get_route_choice_cost()
                        + link.get_toll() / max(EPSILON, vot) * 60
                    )


class Assignment:

    def __init__(self):
        self.agent_types = []
        self.demand_periods = []
        self.demands = []
        # 4-d array
        self.column_pool = {}
        self.network = None
        self.spnetworks = []
        self.accessnetwork = None
        self.map_atstr_id = {}
        self.map_dpstr_id = {}
        self.map_name_atstr = {}
        # number of seconds per simulation interval
        self.simu_rez = 6
        # duration of simulation in minutes
        self.simu_dur = 60
        # simulation start time in minutes
        self.simu_st = 0
        self.has_created_spnet = False

    def update_agent_types(self, at):
        if at.get_type_str() not in self.map_atstr_id:
            self.map_atstr_id[at.get_type_str()] = at.get_id()
        else:
            raise Exception(f'agent type is not unique: {at.get_type_str()}')

        if at.get_name() not in self.map_name_atstr:
            self.map_name_atstr[at.get_name()] = at.get_type_str()
        else:
            raise Exception(f'agent type name is not unique: {at.get_name()}')

        self.agent_types.append(at)

    def update_demand_periods(self, dp):
        if dp.get_period() not in self.map_dpstr_id:
            self.map_dpstr_id[dp.get_period()] = dp.get_id()
        else:
            raise Exception(f'demand period is not unique: {dp.get_period()}')

        self.demand_periods.append(dp)

    def get_agent_type_id(self, at_str):
        try:
            return self.map_atstr_id[at_str]
        except KeyError:
            raise Exception(f'NO agent type: {at_str}')

    def get_demand_period_id(self, dp_str):
        try:
            return self.map_dpstr_id[dp_str]
        except KeyError:
            raise Exception(f'NO demand period: {dp_str}')

    def get_agent_type(self, at_str):
        return self.agent_types[self.get_agent_type_id(at_str)]

    def get_demand_period(self, dp_str):
        return self.demand_periods[self.get_demand_period_id(dp_str)]

    def get_agent_type_str(self, at_id):
        try:
            return self.agent_types[at_id].get_type_str()
        except KeyError:
            raise Exception(f'NO agent type id: {at_id}')

    def get_demand_period_str(self, dp_id):
        try:
            return self.demand_periods[dp_id].get_period()
        except KeyError:
            raise Exception(f'NO demand period id: {dp_id}')

    def update_demands(self, d):
        self.demands.append(d)

    def get_agent_type_count(self):
        return len(self.agent_types)

    def get_demand_period_count(self):
        return len(self.demand_periods)

    def get_agent_types(self):
        return self.agent_types

    def get_demand_periods(self):
        return self.demand_periods

    def get_demands(self):
        return self.demands

    def get_spnetworks(self):
        for sp in self.spnetworks:
            yield sp

    def get_network(self):
        return self.network

    def get_nodes(self):
        """ return list of node objects """
        return self.network.get_nodes()

    def get_links(self):
        """ return list of link objects """
        return self.network.get_links()

    def get_zones(self):
        """ return list of zone IDs """
        return self.network.get_zones()

    def get_column_pool(self):
        return self.column_pool

    def get_column_vec(self, at, dp, orig_zone_id, dest_zone_id):
        return self.column_pool[(at, dp, orig_zone_id, dest_zone_id)]

    def get_agent_orig_node_id(self, agent_id):
        """ return the origin node id of an agent

        exception will be handled by _get_agent() in class Network
        """
        return self.network.get_agent_orig_node_id(agent_id)

    def get_agent_dest_node_id(self, agent_id):
        """ return the destination node id of an agent

        exception will be handled by _get_agent() in class Network
        """
        return self.network.get_agent_dest_node_id(agent_id)

    def get_agent_node_path(self, agent_id, path_only=False):
        """ return the sequence of node IDs along the agent path

        exception will be handled by  _get_agent() in class Network
        """
        return self.network.get_agent_node_path(agent_id, path_only)

    def get_agent_link_path(self, agent_id, path_only=False):
        """ return the sequence of link IDs along the agent path

        exception will be handled by  _get_agent() in class Network
        """
        return self.network.get_agent_link_path(agent_id, path_only)

    def _convert_mode(self, mode):
        """convert mode to the corresponding agent type name and string"""
        if mode in self.map_atstr_id:
            at = self.get_agent_type(mode)
            return at.get_name(), mode

        if mode in self.map_name_atstr:
            return mode, self.map_name_atstr[mode]

        # for distance-based shortest path calculation only
        # it shall not be used with any accessibility evaluations
        if mode.startswith('all'):
            return mode, mode

        raise Exception(f'{mode} is not existing in settings.yml! Please provide a valid mode!')

    def find_path_for_agents(self, mode):
        """ find and set up shortest path for each agent """
        # reset agent type str or mode according to user's input
        at_name, _ = self._convert_mode(mode)
        self.network.set_agent_type_name(at_name)

        find_path_for_agents(self.network, self.column_pool)

    def find_shortest_path(self, from_node_id, to_node_id, mode, seq_type='node'):
        """ call find_shortest_path() from path.py

        exceptions will be handled in find_shortest_path()
        """
        # reset agent type str or mode according to user's input
        at_name, _ = self._convert_mode(mode)
        self.network.set_agent_type_name(at_name)

        # add backward compatibility in case the user still use integer node id's
        from_node_id = str(from_node_id)
        to_node_id = str(to_node_id)

        return find_shortest_path(self.network, from_node_id,
                                  to_node_id, seq_type)

    def benchmark_apsp(self):
        benchmark_apsp(self.network)

    def perform_network_assignment(self, assignment_mode,
                                   iter_num, column_update_num):
        # perform_network_assignment(assignment_mode, iter_num, column_update_num)
        pass

    def perform_network_assignment_DTALite(self, assignment_mode,
                                           iter_num, column_update_num):

        # perform_network_assignment_DTALite(assignment_mode,
        #                                    iter_num,
        #                                    column_update_num)
        pass

    def setup_spnetwork(self):
        if self.has_created_spnet:
            return

        self.network.add_centroids_connectors()
        spvec = {}

        # z is zone id
        for z in self.get_zones():
            if not z:
                continue

            for d in self.demands:
                at = self.get_agent_type(d.get_agent_type_str())
                dp = self.get_demand_period(d.get_period())
                k = (at.get_id(), dp.get_id())
                if k not in spvec:
                    sp = SPNetwork(self.network, at, dp)
                    spvec[k] = sp
                    sp.orig_zones.append(z)
                    self.spnetworks.append(sp)
                else:
                    sp = spvec[k]
                    sp.orig_zones.append(z)

        self.has_created_spnet = True

    def get_link(self, seq_no):
        """ return link object corresponding to link seq no """
        return self.network.get_link(seq_no)

    def get_link_no(self, id):
        """ id is string """
        return self.network.get_link_no(id)

    def get_node_no(self, id):
        """ id is integer """
        return self.network.get_node_no(id)

    def get_agents(self):
        return self.network.get_agents()

    def get_node_label_cost(self, node_no):
        return self.accessnetwork.get_node_label_cost(node_no)

    def get_accessible_nodes(self, source_node_id, time_budget,
                             mode, time_dependent, tau):
        source_node_id = str(source_node_id)
        if source_node_id not in self.network.map_id_to_no:
            raise Exception(f'Node ID: {source_node_id} not in the network')

        assert(time_budget>=0)

        if time_budget == 0:
            return []

        if not self.accessnetwork:
            self.accessnetwork = AccessNetwork(self.network, False)

        # simple caching to avoid duplicate shortest path calculation
        run_sp = False
        if self.accessnetwork.pre_source_node_id != source_node_id:
            self.accessnetwork.set_source_node_id(source_node_id)
            run_sp = True

        at_name, at_str = self._convert_mode(mode)
        if self.accessnetwork.agent_type_name != at_name:
            self.accessnetwork.set_target_mode(at_name)
            at = self.get_agent_type(at_str)
            self.accessnetwork.update_generalized_link_cost(at,
                                                            time_dependent,
                                                            tau)
            run_sp = True

        if run_sp:
            single_source_shortest_path(self.accessnetwork, source_node_id)

        # if max min travel time is less than or equal to time_budget,
        # output the entire node set directly without the following check?
        nodes = []
        for node in self.accessnetwork.get_nodes():
            # do not include the source node itself
            if node.get_node_id() == source_node_id:
                continue

            node_no = node.get_node_no()
            if self.accessnetwork.get_node_label_cost(node_no) <= time_budget:
                nodes.append(node.get_node_id())

        return nodes

    def get_accessible_links(self, source_node_id, time_budget,
                             mode, time_dependent, tau):
        # node id's
        nodes = self.get_accessible_nodes(source_node_id, time_budget,
                                          mode, time_dependent, tau)
        # convert to link id's
        return [
            self.accessnetwork.get_pred_link_id(x) for x in nodes
        ]

    def get_total_simu_intervals(self):
        return ceil(self.simu_dur * 60 / self.simu_rez)

    def have_dep_agents(self, i):
        return self.network.have_dep_agents(i)

    def get_td_agents(self, i):
        return self.network.get_td_agents(i)

    def get_simu_duration(self):
        return self.simu_dur

    def get_simu_start_time(self):
        return self.simu_st

    def initialize_simulation(self, loading_profile):
        profiles = ['constant', 'random', 'uniform']
        if loading_profile not in profiles:
            Warning.warn(
                f'{loading_profile} is not supported!'
                ' constant loading profile will be adopted.'
            )

        agent_id = 1
        links = self.get_links()
        column_pool = self.get_column_pool()

        for k, cv in column_pool.items():
            if cv.get_od_volume() <= 0:
                continue

            # k= (at, dp, orig, dest)
            at = k[0]
            dp = k[1]
            oz = k[2]
            dz = k[3]

            for col in cv.get_columns():
                if col.nodes is None:
                    continue

                # link volume is already set up in UE
                vol = ceil(col.get_volume())
                for j in range(vol):
                    agent = Agent(agent_id, agent_id - 1, at, dp, oz, dz)

                    n = col.get_link_num()
                    agent.curr_link_pos = n - 1
                    agent.link_arr_interval = [-1] * n
                    agent.link_dep_interval = [-1] * n

                    # constant departure time by default
                    t = self.simu_st
                    if loading_profile.startswith('uniform'):
                        t += int(j / col.get_volume() * self.simu_dur)
                    elif loading_profile.startswith('random'):
                        t += randint(0, self.simu_dur - 1)

                    # simulation interval
                    i = self.cast_minute_to_interval(t - self.simu_st)
                    agent.link_arr_interval[-1] = i
                    agent.dep_time = t

                    # set up node path and link path
                    agent.link_path = [x for x in col.links]
                    agent.node_path = [x for x in col.nodes]
                    agent.path_cost = col.get_distance()
                    if i not in self.network.td_agents:
                        self.network.td_agents[i] = []
                    self.network.td_agents[i].append(agent.get_seq_no())
                    self.network.agents.append(agent)

                    agent_id += 1

        # replicate _update_link_travel_time_and_cost()
        for link in links:
            if link.length == 0:
                continue

            # link_capacity is for one hour, i.e., 3600 s
            cap = ceil(link.link_capacity / SECONDS_IN_HOUR * self.simu_rez)
            n1 = self.get_total_simu_intervals()
            n2 = self.get_simu_duration()

            link.outflow_cap = [cap] * n1
            link.cum_arr = [0] * n1
            link.cum_dep = [0] * n1
            # waiting time in terms of simulation interval
            link.waiting_time = [0] * n2

    def get_simu_resolution(self):
        return self.simu_rez

    def get_agent(self, agent_no):
        return self.network._get_agent(agent_no)

    def set_simu_resolution(self, res):
        self.simu_rez = res

    def set_simu_duration(self, dur):
        self.simu_dur = dur

    def set_simu_start_time(self, st):
        self.simu_st = st

    def set_capacity_ratio(self, tau, link_id, r):
        try:
            link_no = self.get_link_no(link_id)
        except KeyError:
            return

        link = self.get_link(link_no)
        link.set_capacity_ratio(tau, r)

    def cast_interval_to_minute(self, i):
        return floor(i * self.simu_rez / SECONDS_IN_MINUTE)

    def cast_interval_to_minute_float(self, i):
        return i * self.simu_rez / SECONDS_IN_MINUTE

    def cast_minute_to_interval(self, m):
        return floor(m * SECONDS_IN_MINUTE / self.simu_rez)


class UI:
    """ an abstract class only with user interfaces """
    def __init__(self, assignment):
        self._base_assignment = assignment

    def get_column_pool(self):
        return self._base_assignment.get_column_pool()

    def get_column_vec(self, at, dp, orig_zone_id, dest_zone_id):
        """ get all columns between two zones given agent type and demand period

        caller is responsible for checking if
        (at, dp, orig_zone_id, dest_zone_id) is in column pool
        """
        self._base_assignment.get_column_vec(at, dp, orig_zone_id, dest_zone_id)

    def get_agent_orig_node_id(self, agent_id):
        """ return the origin node id of an agent """
        return self._base_assignment.get_agent_orig_node_id(agent_id)

    def get_agent_dest_node_id(self, agent_id):
        """ return the destination node id of an agent """
        return self._base_assignment.get_agent_dest_node_id(agent_id)

    def get_agent_node_path(self, agent_id):
        """ return the sequence of node IDs along the agent path """
        return self._base_assignment.get_agent_node_path(agent_id)

    def get_agent_link_path(self, agent_id):
        """ return the sequence of link IDs along the agent path """
        return self._base_assignment.get_agent_link_path(agent_id)

    def get_agent_num(self):
        return self._base_assignment.network.get_agent_count()

    def find_path_for_agents(self, mode='all'):
        """ DEPRECATED

        find and set up shortest path for each agent
        """
        return self._base_assignment.find_path_for_agents(mode)

    def find_shortest_path(self, from_node_id, to_node_id,
                           mode='all', seq_type='node'):
        """ return shortest path between from_node_id and to_node_id

        Parameters
        ----------
        from_node_id
            the starting node id, which shall be a string

        to_node_id
            the ending node id, which shall be a string

        seq_type
            'node' or 'link'. You will get the shortest path in sequence of
            either node IDs or link IDs. The default is 'node'.

        mode
            the target transportation mode which is defined in settings.yml. It
            can be either agent type or its name. For example, 'w' and 'walk'
            are equivalent inputs.

            The default is 'all', which means that links are open to all modes.

        Returns
        -------
        str
            the shortest path between from_node_id and to_node_id.

        Note
        ----
            Exceptions will be thrown if either of from_node_id and and to_node_id
            is not valid node IDs.
        """
        return self._base_assignment.find_shortest_path(
            from_node_id,
            to_node_id,
            mode,
            seq_type
        )

    def get_accessible_nodes(self,
                             source_node_id,
                             time_budget,
                             mode='auto',
                             time_dependent=False,
                             demand_period_id=0):
        """ get the accessible nodes from a node given mode and time budget

        Parameters
        ----------
        source_node_id
            the starting node id for evaluation, which shall be string

        time_budget
            the amount of time to travel in minutes

        mode
            the target transportation mode which is defined in settings.yml. It
            can be either agent type or its name. For example, 'w' and 'walk'
            are equivalent inputs. Its default value is 'a' (i.e., mode auto).

            The default is 'auto'.

        time_dependent
            True or False. Its default value is False.

            If True, the accessibility will be evaluated using the period link
            free-flow travel time (i.e., VDF_fftt). In other words, the
            accessibility is time-dependent.

            If False, the accessibility will be evaluated using the link length
            and the free flow travel speed of each mode.

        demand_period_id
            The sequence number of demand period listed in demand_periods in
            settings.yml. demand_period_id of the first demand_period is 0.

            Use it with time_dependent when there are multiple demand periods.
            Its default value is 0.

        Returns
        -------
        int
            the number of nodes that can be accessible from source_node_id
            given time_budget and mode, and the node list
        """
        nodes = self._base_assignment.get_accessible_nodes(source_node_id,
                                                           time_budget,
                                                           mode,
                                                           time_dependent,
                                                           demand_period_id)

        node_strs = ';'.join(str(x) for x in nodes)

        print(f'number of accessible nodes is {len(nodes)}')
        print(f'accessible nodes are: {node_strs}')

    def get_accessible_links(self,
                             source_node_id,
                             time_budget,
                             mode='auto',
                             time_dependent=False,
                             demand_period_id=0):
        """ get the accessible links from a node given mode and time budget

        Parameters
        ----------
        source_node_id
            the starting node id for evaluation, which shall be string

        time_budget
            the amount of time to travel in minutes

        mode
            the target transportation mode which is defined in settings.yml. It
            can be either agent type or its name. For example, 'w' and 'walk'
            are equivalent inputs.

            The default is 'auto'.

        time_dependent
            True or False. Its default value is False.

            If True, the accessibility will be evaluated using the period link
            free-flow travel time (i.e., VDF_fftt). In other words, the
            accessibility is time-dependent.

            If False, the accessibility will be evaluated using the link length
            and the free flow travel speed of each mode.

        demand_period_id
            The sequence number of demand period listed in demand_periods in
            settings.yml. demand_period_id of the first demand_period is 0.

            Use it with time_dependent when there are multiple demand periods.
            Its default value is 0.

        Returns
        -------
        int
            the number of links that can be accessible from source_node_id
            given time_budget and mode, and the link list
        """
        links = self._base_assignment.get_accessible_links(source_node_id,
                                                           time_budget,
                                                           mode,
                                                           time_dependent,
                                                           demand_period_id)

        link_strs = ';'.join(str(x) for x in links)

        print(f'number of accessible links is {len(links)}')
        print(f'accessible links are: {link_strs}')

    def get_demand_period_str(self, demand_period_id):
        """ return the demand period name given its id

        Parameters
        ----------
        demand_period_id
            The sequence number of demand period listed in demand_periods in
            settings.yml. demand_period_id of the first demand_period is 0.

            Use it with time_dependent when there are multiple demand periods.
            Its default value is 0.

        Returns
        -------
        str
            The name of the corresponding demand period given demand_period_id (
            e.g., 'AM').
        """
        self._base_assignment.get_demand_period_str(demand_period_id)

    def benchmark_apsp(self):
        self._base_assignment.benchmark_apsp()


from os.path import isfile

from path4gmns.colgen import find_ue
from path4gmns.odme import conduct_odme
from path4gmns.io import load_columns, read_demand, read_network, \
                         read_measurements, output_link_performance


def test_odme(sample_data_dir, tmp_output_dir):
    network = read_network(input_dir=sample_data_dir)

    if isfile(tmp_output_dir + '/route_assignment.csv'):
        # bypass perform_column_generation() and call load_columns(network)
        # when there is route_assignment.csv
        load_columns(network, input_dir=tmp_output_dir)
    else:
        read_demand(network, input_dir=sample_data_dir)
        column_gen_num = 20
        column_update_num = 20
        find_ue(network, column_gen_num, column_update_num)

    read_measurements(network, input_dir=sample_data_dir)
    # ODME
    odme_update_num = 20
    conduct_odme(network, odme_update_num)
    # ODME output
    output_link_performance(network, mode='odme', output_dir=tmp_output_dir)


from os import chdir, getcwd

from path4gmns.utils import download_sample_data_sets
from path4gmns.io import read_demand, read_network


def test_download_sample_data_sets(tmp_output_dir):
    orig_dir = getcwd()

    chdir(tmp_output_dir)
    download_sample_data_sets()

    chdir(orig_dir)


def test_data_synthesis(sample_data_dir, tmp_output_dir):
    network = read_network(input_dir=sample_data_dir)

    # try to load synthetic data if there is any. otherwise, synthesize demand
    # and zones, and output them.
    read_demand(network, use_synthetic_data = True, input_dir=tmp_output_dir)


""" global constants """
# for shortest path calculation
MAX_LABEL_COST = 2147483647
EPSILON = 0.00001
# for column generation
MIN_COL_VOL = 0.1
# for accessibility evaluation
MIN_TIME_BUDGET = 10
MAX_TIME_BUDGET = 240
BUDGET_TIME_INTVL = 5
# unit
MILE_TO_METER = 1609
MPH_TO_KPH = 1.609
# simulation
SECONDS_IN_MINUTE = 60
SECONDS_IN_HOUR = 3600


__all__ = ['conduct_odme']


def conduct_odme(ui, odme_update_num, dp_id=0):
    """ calibrate traffic assignment results using traffic observations

    Parameters
    ----------
    ui
        network object generated by pg.read_network()

    odme_update_num
        number of iterations to be performed on ODME

    dp_id
        The sequence number of demand period listed in demand_periods in
        settings.yml. Its default value is 0.

        The current implementation of ODME only supports one demand period.

    Returns
    -------
    None

    Note
    ----
    You will need to call output_columns() and output_link_performance() to
    get the calibrated assignment results (i.e., paths/columns) in
    route_assignment.csv and link volumes (and other link attributes) in
    link_performance.csv.
    """
    print('conduct origin-destination demand matrix estimation (ODME)')

    # base assignment
    A = ui._base_assignment
    # check whether dp_id (demand period id) exists or not
    # exception will be thrown if dp_id does not exist
    A.get_demand_period_str(dp_id)
    # set up SPNetwork
    A.setup_spnetwork()

    # some constants
    delta = 0.05
    step_size = 0.01

    # the current implementation of ODME only supports one demand period
    column_pool = {k: v for k, v in A.get_column_pool().items() if k[1] == dp_id}
    links = A.get_links()
    zones = A.network.zones

    for j in range(odme_update_num):
        _update_link_volume(column_pool, links, zones, dp_id, j)

        # k = (at, dp, oz, dz)
        for k, cv in column_pool.items():
            if cv.is_route_fixed():
                continue

            for col in cv.get_columns():
                vol = col.get_volume()
                if not vol:
                    continue

                path_gradient_cost = 0

                orig_zone = zones[k[2]]
                if orig_zone.prod_obs > 0:
                    if not orig_zone.is_prod_obs_upper_bounded:
                        path_gradient_cost += orig_zone.prod_est_dev
                    elif orig_zone.is_prod_obs_upper_bounded and orig_zone.prod_est_dev > 0:
                        path_gradient_cost += orig_zone.prod_est_dev

                dest_zone = zones[k[3]]
                if dest_zone.attr_obs > 0:
                    if not dest_zone.is_attr_obs_upper_bounded:
                        path_gradient_cost += dest_zone.attr_est_dev
                    elif dest_zone.is_attr_obs_upper_bounded and dest_zone.attr_est_dev > 0:
                        path_gradient_cost += dest_zone.attr_est_dev

                for i in col.links:
                    link = links[i]
                    if link.obs >= 1:
                        if not link.is_obs_upper_bounded:
                            path_gradient_cost += link.est_dev
                        elif link.is_obs_upper_bounded and link.est_dev > 0:
                            path_gradient_cost += link.est_dev

                col.set_gradient_cost(path_gradient_cost)

                change_vol_ub = vol * delta
                change_vol_lb = -change_vol_ub

                change_vol = step_size * path_gradient_cost
                if change_vol < change_vol_lb:
                    change_vol = change_vol_lb
                elif change_vol > change_vol_ub:
                    change_vol = change_vol_ub

                col.set_volume(max(1, vol - change_vol))


def _update_link_volume(column_pool, links, zones, dp_id, iter_num):
    # reset the volume for each link
    for link in links:
        if not link.length:
            break

        link.reset_period_flow_vol()

    # reset the estimated attraction and production
    for z in zones.values():
        z.attr_est = 0
        z.prod_est = 0

    # update estimations and link volume
    # k = (at, dp, oz, dz)
    for k, cv in column_pool.items():
        for col in cv.get_columns():
            vol = col.get_volume()
            if not vol:
                continue

            zones[k[2]].prod_est += vol
            zones[k[3]].attr_est += vol

            for i in col.links:
                # to be consistent with _update_link_and_column_volume()
                # pce_ratio = 1 and vol * pce_ratio
                links[i].increase_period_flow_vol(dp_id, vol)

    # ODME statistics
    total_abs_gap = 0
    total_attr_gap = 0
    total_prod_gap = 0
    total_link_gap = 0

    # calculate estimation deviation for each link
    for link in links:
        if not link.length:
            break

        link.calculate_td_vdf()

        if link.obs < 1:
            continue

        # problematic: est_dev is a scalar rather than a vector?
        # code optimization: wrap it as a member function for class Link
        link.est_dev = link.get_period_flow_vol(dp_id) - link.obs
        total_abs_gap += abs(link.est_dev)
        total_link_gap += link.est_dev / link.obs

    # calculate estimation deviations for each zone
    for z in zones.values():
        if z.attr_obs >= 1:
            dev = z.attr_est - z.attr_obs
            z.attr_est_dev = dev

            total_abs_gap += abs(dev)
            total_attr_gap += dev / z.attr_obs

        if z.prod_obs >= 1:
            dev = z.prod_est - z.prod_obs
            z.prod_est_dev = dev

            total_abs_gap += abs(dev)
            total_prod_gap += dev / z.prod_obs

    print(f'current iteration number in ODME: {iter_num}\n'
          f'total absolute estimation gap: {total_abs_gap:,.2f}\n'
          f'total relative estimation gap (link): {total_link_gap:.2%}\n'
          f'total relative estimation gap (zone attraction): {total_attr_gap:.2%}\n'
          f'total relative estimation gap (zone production): {total_prod_gap:.2%}\n')


from path4gmns.accessibility import evaluate_accessibility, evaluate_equity
from path4gmns.io import read_network


def test_multimodal_accessibility(sample_data_dir, tmp_output_dir):
    network = read_network(input_dir=sample_data_dir)
    # multimodal accessibility evaluation
    evaluate_accessibility(network, output_dir=tmp_output_dir)

    # get accessible nodes and links starting from node 1 with a 5-minute
    # time window for the default mode auto (i.e., 'auto')
    network.get_accessible_nodes(1, 5)
    network.get_accessible_links(1, 5)

    # get accessible nodes and links starting from node 1 with a 15-minute
    # time window for mode walk (i.e., 'w')
    try:
        network.get_accessible_nodes(1, 15, 'w')
        network.get_accessible_links(1, 15, 'w')
    except Exception:
        pass


def test_unimodal_accessibility(sample_data_dir, tmp_output_dir):
    network = read_network(input_dir=sample_data_dir)
    # accessibility evaluation for a target mode only
    evaluate_accessibility(network,
                           single_mode=True,
                           mode='auto',
                           output_dir=tmp_output_dir)


def test_time_dependent_accessibility(sample_data_dir, tmp_output_dir):
    network = read_network(input_dir=sample_data_dir)
    # time-dependent accessibility under the default mode auto
    # for demand period 0 (i.e., VDF_fftt1 in link.csv will be used in the
    # evaluation)
    evaluate_accessibility(network,
                           single_mode=True,
                           time_dependent=True,
                           output_dir=tmp_output_dir)

    # get accessible nodes and links starting from node 1 with a 5-minute
    # time window for the default mode auto for demand period 0
    network.get_accessible_nodes(1, 5, time_dependent=True)
    network.get_accessible_links(1, 5, time_dependent=True)

    # get accessible nodes and links starting from node 1 with a 15-minute
    # time window for mode walk (i.e., 'w') for demand period 0
    try:
        network.get_accessible_nodes(1, 15, 'w', time_dependent=True)
        network.get_accessible_links(1, 15, 'w', time_dependent=True)
    except Exception:
        pass


def test_multimodal_equity(sample_data_dir, tmp_output_dir):
    network = read_network(input_dir=sample_data_dir)
    # multimodal equity evaluation under default time budget (60 min)
    evaluate_equity(network, output_dir=tmp_output_dir)


def test_unimodal_equity(sample_data_dir, tmp_output_dir):
    network = read_network(input_dir=sample_data_dir)
    # equity evaluation for a target mode with time budget as 30 min
    evaluate_equity(network,
                    single_mode=True,
                    mode='auto',
                    time_budget=30,
                    output_dir=tmp_output_dir)


from os.path import isfile

from path4gmns.colgen import find_ue
from path4gmns.simulation import perform_simple_simulation
from path4gmns.io import load_columns, output_agent_trajectory, \
                         read_demand, read_network


def test_simulation(sample_data_dir, tmp_output_dir):
    network = read_network(input_dir=sample_data_dir)

    if isfile(tmp_output_dir + '/route_assignment.csv'):
        # bypass perform_column_generation() and call load_columns(network)
        # when there is route_assignment.csv
        load_columns(network, input_dir=tmp_output_dir)
    else:
        read_demand(network, input_dir=sample_data_dir)
        column_gen_num = 20
        column_update_num = 20
        find_ue(network, column_gen_num, column_update_num)

    # simulation
    perform_simple_simulation(network, 'uniform')
    # simulation output
    output_agent_trajectory(network, output_dir=tmp_output_dir)


from math import ceil


__all__ = ['perform_simple_simulation']


def perform_simple_simulation(ui, loading_profile='uniform'):
    """ perform simple traffic simulation using point queue model

    WARNING
    -------
    The underlying routing decision is made through find_path_for_agents(), which
    must be called before this function.

    Parameters
    ----------
    ui
        network object generated by pg.read_network()

    loading_profile
        demand loading profile, i.e., how agents are loaded to network with respect
        to their departure times. Three loading profiles are supported, which are
        uniform, random, and constant.

        With uniform loading profile, agents will be uniformly distributed between
        [simulation start time, simulation start time + simulation duration).

        With random loading profile, the departure time of each agent is randomly
        set up between [simulation start time, simulation start time + simulation duration).

        With constant loading profile, the departure times of all agents are the
        same, which are simulation start time.

    Returns
    -------
    None

    Note
    ----
    You will need to call output_agent_trajectory() to get the simulation
    results, i.e., trajectory of each agent (in trajectory.csv).
    """
    A = ui._base_assignment
    A.initialize_simulation(loading_profile)
    if not A.get_agents():
        return
    
    print('conduct dynamic traffic assignment (DTA)')

    links = A.get_links()
    nodes = A.get_nodes()

    cum_arr = cum_dep = 0

    # number of simulation intervals in one minute (60s)
    num = A.cast_minute_to_interval(1)
    for i in range(A.get_total_simu_intervals()):
        if i % num == 0:
            print(f'simu time = {i/num} min, CA = {cum_arr}, CD = {cum_dep}')

        # comment out from now on as link.cum_arr and link.cum_dep are not used
        # for critical calculation
        # if i > 0:
        #     for link in links:
        #         link.cum_arr[i] = link.cum_arr[i-1]
        #         link.cum_dep[i] = link.cum_dep[i-1]

        if A.have_dep_agents(i):
            for a_no in A.get_td_agents(i):
                a = A.get_agent(a_no)
                if a.link_path is None:
                    continue

                # retrieve the first link given link path is in reverse order
                link_no = a.link_path[-1]
                link = links[link_no]
                # link.cum_arr[i] += 1
                link.entr_queue.append(a_no)
                cum_arr += 1

        for link in links:
            while link.entr_queue:
                a_no = link.entr_queue.popleft()
                agent = A.get_agent(a_no)
                link.exit_queue.append(a_no)
                intvl = A.cast_minute_to_interval(link.get_period_fftt(0))
                agent.update_dep_interval(intvl)

        for node in nodes:
            m = node.get_incoming_link_num()
            for j in range(m):
                # randomly select the first link
                pos = (i + j) % m
                link = node.incoming_links[pos]

                while link.outflow_cap[i] and link.exit_queue:
                    a_no = link.exit_queue[0]
                    agent = A.get_agent(a_no)

                    if agent.get_curr_dep_interval() > i:
                        break

                    if agent.reached_last_link():
                        # link.cum_dep[i] += 1
                        cum_dep += 1
                    else:
                        link_no = agent.get_next_link_no()
                        next_link = links[link_no]
                        next_link.entr_queue.append(a_no)
                        agent.set_dep_interval(i)
                        # set up arrival time for the next link, i.e., next_link
                        agent.set_arr_interval(i, 1)

                        travel_intvl = i - agent.get_arr_interval()
                        waiting_intvl = travel_intvl - A.cast_minute_to_interval(link.get_period_fftt(0))
                        # arrival time in minutes
                        arr_minute = A.cast_interval_to_minute(agent.get_arr_interval())
                        link.update_waiting_time(arr_minute, waiting_intvl)
                        # link.cum_dep[i] += 1
                        # next_link.cum_arr[i] += 1

                    agent.increment_link_pos()
                    # remove agent from exit queue
                    link.exit_queue.popleft()
                    link.outflow_cap[i] -= 1


from os.path import isfile
from random import randint

from path4gmns.io import output_agent_paths, read_demand, read_network


def test_routing_engine(sample_data_dir):
    network = read_network(input_dir=sample_data_dir)
    network.benchmark_apsp()


def test_find_shortest_path(sample_data_dir):
    network = read_network(input_dir=sample_data_dir)

    # shortest path (node id) from node 1 to node 2
    network.find_shortest_path(1, 2)
    # shortest path (link id) from node 1 to node 2
    network.find_shortest_path(1, 2, seq_type='link')

    # retrieve the shortest path under a specific mode (which must be defined
    # in settings.yaml)
    if isfile(sample_data_dir + '/settings.yml'):
        # shortest path (node id) from node 1 to node 2
        network.find_shortest_path(1, 2, mode='a')
        # shortest path (link id) from node 1 to node 2
        network.find_shortest_path(1, 2, mode='a', seq_type='link')


def test_find_shortest_path_for_agents(sample_data_dir, tmp_output_dir):
    """ find_path_for_agents has been DEPRECATED """
    network = read_network(input_dir=sample_data_dir)
    read_demand(network, input_dir=sample_data_dir)

    # find agent paths under a specific mode defined in settings.yaml,
    # say, a (i.e., auto)
    # network.find_path_for_agents('a') or network.find_path_for_agents('auto')
    network.find_path_for_agents()

    if network.get_agent_num() == 0:
        return

    agent_id = randint(0, network.get_agent_num() - 1)
    # origin node id of agent
    network.get_agent_orig_node_id(agent_id)
    # destination node id of agent
    network.get_agent_dest_node_id(agent_id)
    # shortest path (node id) of agent
    network.get_agent_node_path(agent_id)
    # shortest path (link id) of agent
    network.get_agent_link_path(agent_id)

    # output unique agent paths to a csv file
    output_agent_paths(network, output_dir=tmp_output_dir)
    # exclude geometry info from the output file
    output_agent_paths(network, False, output_dir=tmp_output_dir)


import os
import ctypes
import platform
from multiprocessing import Process
from time import sleep


__all__ = ['perform_network_assignment_DTALite', 'run_DTALite']


_os = platform.system()
if _os.startswith('Windows'):
    _dtalite_dll = os.path.join(os.path.dirname(__file__), 'bin/DTALite.dll')
    _dtalitemm_dll = os.path.join(os.path.dirname(__file__), 'bin/DTALiteMM.dll')
elif _os.startswith('Linux'):
    _dtalite_dll = os.path.join(os.path.dirname(__file__), 'bin/DTALite.so')
    _dtalitemm_dll = os.path.join(os.path.dirname(__file__), 'bin/DTALiteMM.so')
elif _os.startswith('Darwin'):
    # check CPU is Intel or Apple Silicon
    if platform.machine().startswith('x86_64'):
        _dtalite_dll = os.path.join(os.path.dirname(__file__), 'bin/DTALite_x86.dylib')
        _dtalitemm_dll = os.path.join(os.path.dirname(__file__), 'bin/DTALiteMM_x86.dylib')
    else:
        _dtalite_dll = os.path.join(os.path.dirname(__file__), 'bin/DTALite_arm.dylib')
        _dtalitemm_dll = os.path.join(os.path.dirname(__file__), 'bin/DTALiteMM_arm.dylib')
else:
    raise Exception('Please build the shared library compatible to your OS\
                    using source files')


def _emit_log(log_file='log_main.txt'):
    with open(log_file, 'r') as fp:
        for line in fp:
            print(line)


def perform_network_assignment_DTALite(assignment_mode,
                                       column_gen_num,
                                       column_update_num):
    """ DEPRECATED Python interface to call DTALite (precompiled as shared library)

    perform network assignment using the selected assignment mode

    WARNING
    -------
    MAKE SURE TO BACKUP route_assignment.csv and link_performance.csv if you have
    called perform_network_assignment() before. Otherwise, they will be
    overwritten by results generated by DTALite.

    Parameters
    ----------
    assignment_mode
        0: Link-based UE, only generates link performance file without agent path file

        1: Path-based UE, generates link performance file and agent path file

        2: UE + dynamic traffic assignment (DTA), generates link performance file and agent path file

        3: ODME

    column_gen_num
        number of iterations to be performed before on generating column pool

    column_update_iter
        number of iterations to be performed on optimizing column pool

    Returns
    -------
    None

    Note
    ----
    The output will depend on the selected assignment_mode.

        Link-based UE: link_performance.csv

        Path-based UE: route_assignment.csv and link_performance.csv

        UE + DTA: route_assignment.csv and link_performance.csv

    route_assignment.csv: paths/columns

    link_performance.csv: assigned volumes and other link attributes
    on each link
    """
    # make sure assignment_mode is right
    assert(assignment_mode in [0, 1, 2, 3])
    # make sure iteration numbers are both non-negative
    assert(column_gen_num>=0)
    assert(column_update_num>=0)

    _dtalite_engine = ctypes.cdll.LoadLibrary(_dtalite_dll)
    _dtalite_engine.network_assignment.argtypes = [ctypes.c_int,
                                                   ctypes.c_int,
                                                   ctypes.c_int]

    print('This function has been deprecated, and will be removed later!'
          'Please use run_DTALite() instead!')
    print('\nDTALite run starts\n')

    if _os.startswith('Windows'):
        _dtalite_engine.network_assignment(assignment_mode,
                                           column_gen_num,
                                           column_update_num)

        _emit_log()

        print('\nDTALite run completes\n')
        print(
            f'check link_performance.csv in {os.getcwd()} for link performance\n'
            f'check route_assignment.csv in {os.getcwd()} for unique agent paths\n'
        )
    else:
        # the following multiprocessing call does not work for Windows,
        # and there is no solution.
        # OSError: [WinError 87] The parameter is incorrect
        proc_dta = Process(
            target=_dtalite_engine.network_assignment,
            args=(assignment_mode, column_gen_num, column_update_num,)
        )

        proc_print = Process(target=_emit_log)

        proc_dta.start()
        proc_dta.join()

        if proc_dta.exitcode is not None:
            sleep(0.1)
            proc_print.start()
            proc_print.join()
            if proc_dta.exitcode == 0:
                print('DTALite run completes!\n')
                print(
                    f'check link_performance.csv in {os.getcwd()} for link performance\n'
                    f'check route_assignment.csv in {os.getcwd()} for unique agent paths\n'
                )
            else:
                print('DTALite run terminates!')


def run_DTALite():
    """ Python interface to call the latest DTALite

    This version of DTALite includes all-new Logbook, enhanced scenario handling,
    improved I/O functionality, and so on.

    Its source code can be found at https://github.com/asu-trans-ai-lab/DTALite/tree/feature/multimodal.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Note
    ----
    It is NOT compatible with the classic DTALite (i.e., perform_network_assignment_DTALite()).

    Only use the following data set from
    https://github.com/asu-trans-ai-lab/DTALite/tree/feature/multimodal/data.

    Please run the script calling this API using system terminal rather than
    Python console for proper logging.
    """
    _dtalitemm_engine = ctypes.cdll.LoadLibrary(_dtalitemm_dll)
    print('\nDTALite run starts\n')

    proc_dta = Process(target=_dtalitemm_engine.DTALiteAPI())
    proc_print = Process(target=_emit_log, args=('log_DTA.txt',))

    proc_dta.start()
    proc_dta.join()

    if proc_dta.exitcode is not None:
        sleep(0.1)
        proc_print.start()
        proc_print.join()
        if proc_dta.exitcode == 0:
            print('DTALite run completes!')
        else:
            print('DTALite run terminates!')


import csv
import os
import warnings

from .classes import Node, Link, Zone, Network, Column, ColumnVec, VDFPeriod, \
                     AgentType, DemandPeriod, Demand, SpecialEvent, Assignment, UI

from .colgen import update_links_using_columns
from .consts import EPSILON, MILE_TO_METER, MPH_TO_KPH
from .utils import InvalidRecord, _are_od_connected, _convert_boundaries, \
                   _convert_str_to_float, _convert_str_to_int, _get_time_stamp, \
                   _update_dest_zone, _update_orig_zone
from .zonesyn import network_to_zones


__all__ = [
    'read_network',
    'read_demand',
    'read_measurements',
    'load_demand',
    'load_columns',
    'output_columns',
    'output_link_performance',
    'output_agent_paths',
    'output_agent_trajectory',
    'output_synthetic_zones',
    'output_synthetic_demand'
]


def read_nodes(input_dir,
               nodes,
               map_id_to_no,
               map_no_to_id,
               zones,
               load_demand):

    """ step 1: read input_node """
    with open(input_dir+'/node.csv', 'r') as fp:
        print('read node.csv')

        reader = csv.DictReader(fp)
        node_no = 0
        for line in reader:
            # set up node_id
            node_id = line['node_id']
            # node_id should be unique
            if node_id in map_id_to_no:
                print(f'Duplicate node id found {node_id}. Record discarded!')
                continue

            # set up zone_id
            try:
                zone_id = line['zone_id']
            except KeyError:
                zone_id = ''

            # treat them as string
            coord_x = line['x_coord']
            coord_y = line['y_coord']

            # activity node
            is_activity_node = False
            try:
                b = _convert_str_to_int(line['boundary'])
                if b:
                    is_activity_node = True
            except (KeyError, InvalidRecord):
                pass

            # construct node object
            node = Node(node_no, node_id, zone_id, coord_x, coord_y, is_activity_node)
            nodes.append(node)

            # set up mapping between node_no and node_id
            map_id_to_no[node_id] = node_no
            map_no_to_id[node_no] = node_id

            # bin_index for equity evaluation
            try:
                bin_index = _convert_str_to_int(line['bin_index'])
            except (KeyError, InvalidRecord):
                bin_index = 0

            # associate node_id with corresponding zone
            if zone_id not in zones:
                # only take the value of bin_index from the first node
                # associated with each zone
                z = Zone(zone_id, bin_index)
                zones[zone_id] = z

            zones[zone_id].add_node(node_id)
            if is_activity_node:
                zones[zone_id].add_activity_node(node_id)

            node_no += 1

        print(f'the number of nodes is {node_no:,d}')

        if load_demand:
            zone_size = len(zones)
            if '' in zones:
                zone_size -= 1

            if zone_size == 0:
                raise Exception('there are NO VALID zones from node.csv')

            print(f'the number of zones is {zone_size:,d}\n')


def read_links(input_dir,
               links,
               nodes,
               map_id_to_no,
               link_ids,
               demand_period_size,
               length_unit,
               speed_unit,
               load_demand):
    """ step 2: read input_link """
    with open(input_dir+'/link.csv', 'r') as fp:
        print('read link.csv')

        reader = csv.DictReader(fp)
        link_no = 0
        # a temporary container (set) to validate the uniqueness of a link id
        link_ids_ = set()
        for line in reader:
            # link id shall be unique
            link_id = line['link_id']
            # binary search shall be fast enough
            if link_id in link_ids_:
                print(f'Duplicate link id found {link_id}. Record discarded!')
                continue

            # validity check
            from_node_id = line['from_node_id']
            to_node_id = line['to_node_id']

            try:
                from_node_no = map_id_to_no[from_node_id]
            except KeyError:
                print(
                    f'Exception: Node ID {from_node_id}'
                    ' NOT in the network!!'
                )
                continue

            try:
                to_node_no = map_id_to_no[to_node_id]
            except KeyError:
                print(
                    f'Exception: Node ID {to_node_id}'
                    ' NOT in the network!!'
                )
                continue

            try:
                length = _convert_str_to_float(line['length'])
            except InvalidRecord:
                continue

            # pass validity check

            # for the following attributes,
            # if they are not None, convert them to the corresponding types
            # if they are None's, set them using the default values
            try:
                lanes = _convert_str_to_int(line['lanes'])
            except InvalidRecord:
                lanes = 1

            try:
                link_type = _convert_str_to_int(line['link_type'])
            except (KeyError, InvalidRecord):
                link_type = 1

            try:
                free_speed = _convert_str_to_float(line['free_speed'])
            except InvalidRecord:
                free_speed = 60

            # issue: int??
            try:
                capacity = _convert_str_to_int(line['capacity'])
            except (KeyError, InvalidRecord):
                capacity = 1999

            try:
                toll = _convert_str_to_float(line['toll'])
            except (KeyError, InvalidRecord):
                toll = 0

            # if link.csv does not have no column 'allowed_uses',
            # set allowed_uses to 'all'
            # developer's note:
            # we may need to change this implementation as we cannot deal with
            # cases a link which is not open to any modes
            try:
                allowed_uses = line['allowed_uses']
                if not allowed_uses:
                    raise InvalidRecord
            except (KeyError, InvalidRecord):
                allowed_uses = 'all'

            # if link.csv does not have no column 'geometry',
            # set geometry to ''
            try:
                geometry = line['geometry']
            except KeyError:
                geometry = ''

            link_ids[link_id] = link_no

            # unit conversion
            if length_unit.startswith('meter') or length_unit == 'm':
                length = length / MILE_TO_METER
            elif length_unit.startswith('kilometer') or length_unit.startswith('km'):
                length = length / MPH_TO_KPH

            if speed_unit.startswith('kmh') or speed_unit.startswith('kph'):
                free_speed = free_speed / MPH_TO_KPH

            # construct link object
            link = Link(link_id,
                        link_no,
                        from_node_no,
                        to_node_no,
                        from_node_id,
                        to_node_id,
                        length,
                        lanes,
                        link_type,
                        free_speed,
                        capacity,
                        toll,
                        allowed_uses,
                        geometry,
                        demand_period_size)

            # VDF Attributes
            for i in range(demand_period_size):
                dp_id_str = str(i+1)
                header_vdf_alpha = 'VDF_alpha' + dp_id_str
                header_vdf_beta = 'VDF_beta' + dp_id_str
                header_vdf_mu = 'VDF_mu' + dp_id_str
                header_vdf_fftt = 'VDF_fftt' + dp_id_str
                header_vdf_cap = 'VDF_cap' + dp_id_str
                header_vdf_phf = 'VDF_phf' + dp_id_str

                # case i: link.csv does not VDF attributes at all
                # case ii: link.csv only has partial VDF attributes
                # under case i, we will set up only one VDFPeriod object using
                # default values
                # under case ii, we will set up some VDFPeriod objects up to
                # the number of complete set of VDF_alpha, VDF_beta, and VDF_mu
                try:
                    VDF_alpha = _convert_str_to_float(line[header_vdf_alpha])
                except (KeyError, InvalidRecord):
                    if i == 0:
                        # default value will be applied in the constructor
                        VDF_alpha = 0.15
                    else:
                        break

                try:
                    VDF_beta = _convert_str_to_float(line[header_vdf_beta])
                except (KeyError, InvalidRecord):
                    if i == 0:
                        # default value will be applied in the constructor
                        VDF_beta = 4
                    else:
                        break

                try:
                    VDF_mu = _convert_str_to_float(line[header_vdf_mu])
                except (KeyError, InvalidRecord):
                    if i == 0:
                        # default value will be applied in the constructor
                        VDF_mu = 1000
                    else:
                        break

                try:
                    VDF_fftt = _convert_str_to_float(line[header_vdf_fftt])
                except (KeyError, InvalidRecord):
                    # set it up using length and free_speed from link
                    VDF_fftt = length / max(EPSILON, free_speed) * 60

                try:
                    VDF_cap = _convert_str_to_float(line[header_vdf_cap])
                except (KeyError, InvalidRecord):
                    # set it up using capacity from link
                    VDF_cap = capacity * lanes

                # not a mandatory column
                try:
                    VDF_phf = _convert_str_to_float(line[header_vdf_phf])
                except (KeyError, InvalidRecord):
                    # default value will be applied in the constructor
                    VDF_phf = -1

                # construct VDFPeriod object
                vdf = VDFPeriod(i, VDF_alpha, VDF_beta, VDF_mu,
                                VDF_fftt, VDF_cap, VDF_phf)

                link.vdfperiods.append(vdf)

            # set up outgoing links and incoming links
            from_node = nodes[from_node_no]
            to_node = nodes[to_node_no]
            from_node.add_outgoing_link(link)
            to_node.add_incoming_link(link)
            links.append(link)

            # set up zone degrees
            if load_demand:
                oz_id = from_node.get_zone_id()
                dz_id = to_node.get_zone_id()
                _update_orig_zone(oz_id)
                _update_dest_zone(dz_id)

            link_no += 1
            link_ids_.add(link_id)

        print(f'the number of links is {link_no:,d}\n')


def _read_demand(input_dir,
                 file,
                 agent_type_id,
                 demand_period_id,
                 zones,
                 column_pool,
                 check_connectivity=False):
    """ step 3:read input_agent """
    with open(input_dir+'/'+file, 'r') as fp:
        print('read '+file)

        at = agent_type_id
        dp = demand_period_id

        reader = csv.DictReader(fp)
        valid_vol = 0
        invalid_vol = 0
        invalid_od_num = 0
        for line in reader:
            oz_id = line['o_zone_id']
            # o_zone_id does not exist in node.csv, discard it
            if oz_id not in zones:
                continue

            dz_id = line['d_zone_id']
            # d_zone_id does not exist in node.csv, discard it
            if dz_id not in zones:
                continue

            try:
                vol = _convert_str_to_float(line['volume'])
            except InvalidRecord:
                continue

            # discard invalid OD pair, case I: invalid volume
            if vol <= 0:
                invalid_od_num += 1
                continue

            # discard invalid OD pair, case II: O and D are the same
            if oz_id == dz_id:
                invalid_od_num += 1
                invalid_vol += vol
                continue

            # precheck on connectivity of each OD pair
            if check_connectivity and not (_are_od_connected(oz_id, dz_id)):
                continue

            # set up volume for ColumnVec
            if (at, dp, oz_id, dz_id) not in column_pool:
                column_pool[(at, dp, oz_id, dz_id)] = ColumnVec()
            column_pool[(at, dp, oz_id, dz_id)].increase_volume(vol)

            valid_vol += vol

        print(
            f'the total valid demand is {valid_vol:,.3f}\n'
            f'{invalid_od_num:,d} invalid OD pairs are found. '
            f'Total discarded volume: {invalid_vol:,.2f}\n'
        )

        if valid_vol == 0:
            # there are four cases.
            # Case I:   zones is empty
            #           This shall be caught by read_node().
            # Case II:  zones is not empty and invalid_od_num == 0
            #           Every zone_id present in demand.csv is not found in
            #           node.csv. This implies data inconsistency between these
            #           two files on zone_id). Another possibility is that volume
            #           is not numerical.
            # Case III: zones is not empty, invalid_od_num > 0, and invalid_vol == 0
            #           All OD pairs in demand.csv have invalid volume (i.e.,
            #           volume is zero or less).
            # Case IV:  zones is not empty, invalid_od_num > 0, and invalid_vol > 0
            #           This implies either O and D are the same for every OD
            #           pair or no OD pair is not connected.
            if invalid_od_num == 0:
                message = 'volume is not encoded as numerical or Every zone id ' \
                          'present in demand.csv is not found in node.csv.\n\n' \
                          'For the latter one, different zone id representations ' \
                          'could be the reason. For example, zone id is encoded ' \
                          'as integer in demand.csv but decimal in node.csv. ' \
                          'Note that zone_id CANNOT be decimal per GMNS specification!\n'\
                          'Hint: Use an advanced text editor (not Excel) to verify. '
            elif invalid_od_num > 0 and invalid_vol == 0:
                message = 'At lease one OD pair shall have positive volume!\n'
            else:
                message = 'No connected OD pairs are found (i.e., no valid paths ' \
                          'between O and D for each OD pair) or Every OD pair ' \
                          'have the same O and D!\n'

            raise Exception(message)


def load_demand(ui,
                agent_type_str='a',
                demand_period_str='AM',
                input_dir='.',
                filename='demand.csv'):
    """
    load demand for an agent type and a demand period

    this is an user interface while _read_demand() is intended for internal use.
    """
    A = ui._base_assignment

    at = A.get_agent_type_id(agent_type_str)
    dp = A.get_demand_period_id(demand_period_str)
    # do not check connectivity of OD pairs
    _read_demand(input_dir, filename, at, dp, A.network.zones, A.column_pool, False)


def _read_zones(ui, input_dir='.', filename='syn_zone.csv'):
    """ read syn_zone.csv to set up (synthetic) zones """
    with open(input_dir+'/'+filename, 'r') as fp:
        print('read syn_zone.csv')

        A = ui._base_assignment
        zones = A.network.zones
        zones.clear()

        node_objs = A.get_nodes()

        reader = csv.DictReader(fp)
        for line in reader:
            zone_id = line['zone_id']
            if not zone_id:
                continue

            nodes = line['activity_nodes']
            if not nodes:
                continue

            # just in case that there is empty space in between ';'
            node_ids = [x.strip() for x in nodes.split(';') if x.strip()]
            if not node_ids:
                continue

            try:
                bin_index = _convert_str_to_int(line['bin_index'])
            except (KeyError, InvalidRecord):
                bin_index = 0

            try:
                x = _convert_str_to_float(line['x_coord'])
                y = _convert_str_to_float(line['y_coord'])
            except (KeyError, InvalidRecord):
                x = 91
                y = 181

            try:
                U, D, L, R = _convert_boundaries(line['geometry'])
            except (KeyError, InvalidRecord):
                U = D = L = R = ''

            try:
                prod = _convert_str_to_int(line['production'])
            except (KeyError, InvalidRecord):
                prod = 0

            if zone_id not in zones:
                z = Zone(zone_id, bin_index)
                z.activity_nodes = [x for x in node_ids]
                z.nodes = [x for x in node_ids]
                z.set_coord(x, y)
                z.set_geo(U, D, L, R)
                z.set_production(prod)
                zones[zone_id] = z
                # update zone info for each node
                for id in node_ids:
                    try:
                        no = A.get_node_no(id)
                        node_objs[no].zone_id = zone_id
                    except (IndexError, KeyError):
                        continue
            else:
                raise Exception(f'DUPLICATE zone id: {zone_id}')

        print(f'the number of zones is {len(zones):,d}')


def read_demand_matrix(input_dir, agent_type_id, demand_period_id,
                       zone_to_node_dict, column_pool):
    """ read demand matrix from input_matrix

    not in use
    """
    with open(input_dir+'/input_matrix.csv', 'r') as fp:
        print('read input_matrix.csv')

        at = agent_type_id
        dp = demand_period_id

        total_vol = 0
        reader = csv.DictReader(fp)
        for line in reader:
            oz_id = line['od']
            # o_zone_id does not exist in node.csv, discard it
            if oz_id not in zone_to_node_dict:
                continue

            for dz_str, vol_str in line:
                dz_id = _convert_str_to_int(dz_str)
                if dz_id == oz_id:
                    continue

                # d_zone_id does not exist in node.csv, discard it
                if dz_id not in zone_to_node_dict:
                    continue

                try:
                    vol = _convert_str_to_float(vol_str)
                except InvalidRecord:
                    continue

                if vol == 0:
                    continue

                if not _are_od_connected(oz_id, dz_id):
                    continue

                if (at, dp, oz_id, dz_id) not in column_pool:
                    column_pool[(at, dp, oz_id, dz_id)] = ColumnVec()
                    column_pool[(at, dp, oz_id, dz_id)].set_volume(vol)
                else:
                    raise Exception(
                        f'DUPLICATE OD pair found between {oz_id} and {dz_id}'
                    )

                total_vol += vol

            print(f'the number of agents is {total_vol}')

            if total_vol == 0:
                raise Exception(
                    'NO VALID OD VOLUME!! DOUBLE CHECK YOUR input_matrix.csv'
                )


def _auto_setup(assignment):
    """ automatically set up one demand period and one agent type

    The two objects will be set up using the default constructors using the
    default values. See class DemandPeriod and class AgentType for details
    """
    at = AgentType()
    dp = DemandPeriod()
    d = Demand()

    assignment.update_agent_types(at)
    assignment.update_demand_periods(dp)
    assignment.update_demands(d)


def read_settings(input_dir, assignment):
    try:
        import yaml as ym

        with open(input_dir+'/settings.yml') as file:
            print('read settings.yml\n')

            settings = ym.full_load(file)

            # agent types
            agents = settings['agents']
            for i, a in enumerate(agents):
                agent_type = a['type']
                agent_name = a['name']
                # possible duplication check
                if agent_type in assignment.map_atstr_id:
                    warnings.warn(f'duplicate agent type found: {agent_type}')
                    continue

                agent_vot = a['vot']
                agent_flow_type = a['flow_type']
                agent_pce = a['pce']
                agent_ffs = a['free_speed']

                try:
                    agent_use_link_ffs = a['use_link_ffs']
                except KeyError:
                    agent_use_link_ffs = True

                at = AgentType(i,
                               agent_type,
                               agent_name,
                               agent_vot,
                               agent_flow_type,
                               agent_pce,
                               agent_ffs,
                               agent_use_link_ffs)

                assignment.update_agent_types(at)

            # add the default mode if it does not exist
            if AgentType.get_default_type_str() not in assignment.map_atstr_id:
                assignment.update_agent_types(AgentType())

            # demand periods
            demand_periods = settings['demand_periods']
            for i, d in enumerate(demand_periods):
                period = d['period']
                time_period = d['time_period']

                dp = DemandPeriod(i, period, time_period)
                # special event
                try:
                    s = d['special_event']
                    enable = s['enable']
                    # no need to set up a special event if it is off
                    if not enable:
                        raise KeyError

                    name = s['name']
                    se = SpecialEvent(name)

                    links = s['affected_links']
                    for link in links:
                        link_id = str(link['link_id'])
                        ratio = link['capacity_ratio']
                        se.affected_links[link_id] = ratio

                    dp.special_event = se
                except KeyError:
                    pass

                assignment.update_demand_periods(dp)

            # demand files
            demands = settings['demand_files']
            for i, d in enumerate(demands):
                demand_file = d['file_name']
                demand_period = d['period']
                demand_type = d['agent_type']

                if demand_type not in assignment.map_atstr_id:
                    raise Exception(
                        f'{demand_type} is not found as an entry of agents in settings.yml'
                    )

                demand = Demand(i, demand_period, demand_type, demand_file)
                assignment.update_demands(demand)

            # simulation setup
            try:
                simulation = settings['simulation']
                # simulation resolution
                res = simulation['resolution']
                assert(int(res)>1)
                assignment.set_simu_resolution(int(res))
                # simulation timings
                dp_str = simulation['period']
                dp = assignment.get_demand_period(dp_str)
                st = dp.get_start_time()
                dur = dp.get_duration()
                assignment.set_simu_start_time(st)
                assignment.set_simu_duration(dur)
            except KeyError:
                pass

    except ImportError:
        # just in case user does not have pyyaml installed
        # warnings.warn(
        #     'Please install pyyaml next time!\n'
        #     'Engine will set up one demand period and one agent type using '
        #     'default values for you, which might NOT reflect your case!'
        # )
        _auto_setup(assignment)
    except FileNotFoundError:
        # just in case user does not provide settings.yml
        # warnings.warn(
        #     'Please provide settings.yml next time!\n'
        #     'Engine will set up one demand period and one agent type using '
        #     'default values for you, which might NOT reflect your case!'
        # )
        _auto_setup(assignment)
    except Exception as e:
        raise e


def read_network(length_unit='mile', speed_unit='mph', input_dir='.'):
    len_units = ['kilometer', 'km', 'meter', 'm', 'mile', 'mi']
    spd_units = ['kmh', 'kph', 'mph']

    # length and speed units check
    # linear search is OK for such small lists
    if length_unit not in len_units:
        units = ', '.join(len_units)
        raise Exception(
            f'Invalid length unit: {length_unit} !'
            f' Please choose one available unit from {units}'
        )

    if speed_unit not in spd_units:
        units = ', '.join(spd_units)
        raise Exception(
            f'Invalid speed unit: {speed_unit} !'
            f' Please choose one available unit from {units}'
        )

    assignm = Assignment()
    network = Network()

    read_settings(input_dir, assignm)

    read_nodes(input_dir,
               network.nodes,
               network.map_id_to_no,
               network.map_no_to_id,
               network.zones,
               load_demand)

    read_links(input_dir,
               network.links,
               network.nodes,
               network.map_id_to_no,
               network.link_ids,
               assignm.get_demand_period_count(),
               length_unit,
               speed_unit,
               load_demand)

    network.update()
    assignm.network = network

    return UI(assignm)


def load_columns(ui, input_dir='.'):
    with open(input_dir+'/route_assignment.csv', 'r') as f:
        print('read route_assignment.csv')

        A = ui._base_assignment
        cp = A.get_column_pool()

        # just in case agent_id was not output
        last_agent_id = 0
        reader = csv.DictReader(f)
        for line in reader:
            # critical info
            oz_id = line['o_zone_id']
            dz_id = line['d_zone_id']

            try:
                vol = _convert_str_to_float(line['volume'])
            except InvalidRecord:
                continue

            # skip zero-volume column
            if not vol:
                continue

            node_seq = line['node_sequence']
            if not node_seq:
                continue

            link_seq = line['link_sequence']
            if not link_seq:
                continue

            # non-critical info
            try:
                agent_id = _convert_str_to_int(line['agent_id'])
            except InvalidRecord:
                agent_id = last_agent_id + 1

            last_agent_id = agent_id

            # it could be empty
            # path_id = line['path_id']

            at = line['agent_type']
            if not at:
                continue
            else:
                # back-compatible on 'p' and 'passenger'
                try:
                    at = A.get_agent_type_id(at)
                except Exception:
                    # replace 'p' with 'a'
                    if at.startswith(AgentType.get_legacy_type_str()):
                        at = A.get_agent_type_id(AgentType.get_default_type_str())
                    else:
                        warnings.warn(
                            f'agent_type {at} is not existing in settings.yml.'
                            'this record is discarded'
                        )
                        continue

            dp = line['demand_period']
            if not dp:
                continue
            else:
                dp = A.get_demand_period_id(dp)

            try:
                toll = _convert_str_to_float(line['toll'])
            except InvalidRecord:
                toll = 0

            try:
                tt = _convert_str_to_float(line['travel_time'])
            except InvalidRecord:
                tt = 0

            try:
                dist = _convert_str_to_float(line['distance'])
            except InvalidRecord:
                dist = 0

            # it could be empty
            geo = line['geometry']

            if (at, dp, oz_id, dz_id) not in cp:
                cp[(at, dp, oz_id, dz_id)] = ColumnVec()

            cv = A.get_column_vec(at, dp, oz_id, dz_id)
            path_id = cv.get_column_num()
            col = Column(path_id)

            try:
                col.nodes = [A.get_node_no(x) for x in reversed(node_seq.split(';')) if x]
            except KeyError:
                raise Exception(
                    'Invalid node found on column!!'
                    'Did you use route_assignment.csv from a different network?'
                )

            try:
                # if x is only needed for columns generated from DTALite,
                # which have the trailing ';' and leads to '' after split
                col.links = [
                    A.get_link_no(x) for x in reversed(link_seq.split(';')) if x
                ]
            except KeyError:
                raise Exception(
                    'INVALID link found on column!!'
                    'Did you use route_assignment.csv from a different network?'
                )
            except ValueError:
                raise Exception(
                    f'INVALID LINK PATH found for agent id: {agent_id}'
                )

            # the following four are non-critical info
            col.set_volume(vol)
            col.set_toll(toll)
            col.set_travel_time(tt)
            col.set_geometry(geo)

            # deprecate node_sum and adopt the same implementation in colgen.py
            existing = False

            if dist == 0:
                sum(A.get_link(x).get_length() for x in col.links)

            for col_ in cv.get_columns():
                if col_.get_distance() != dist:
                    continue

                if col_.get_links() == col.links:
                    col_.increase_volume(vol)
                    existing = True
                    break

            if not existing:
                col.set_distance(dist)
                cv.add_new_column(col)
                cv.increase_volume(vol)

        update_links_using_columns(ui)


def output_columns(ui, output_geometry=True, output_dir='.'):
    with open(output_dir+'/route_assignment.csv', 'w',  newline='') as fp:
        base = ui._base_assignment

        nodes = base.get_nodes()
        links = base.get_links()
        column_pool = base.get_column_pool()

        writer = csv.writer(fp)

        line = ['agent_id',
                'o_zone_id',
                'd_zone_id',
                'path_id',
                'agent_type',
                'demand_period',
                'volume',
                'toll',
                'travel_time',
                'distance',
                'node_sequence',
                'link_sequence',
                'geometry']

        writer.writerow(line)

        path_sep = ';'
        i = 0
        for k, cv in column_pool.items():
            # k = (at_id, dp_id, oz_id, dz_id)
            at_id = k[0]
            dp_id = k[1]
            oz_id = k[2]
            dz_id = k[3]

            at_str = base.get_agent_type_str(at_id)
            dp_str = base.get_demand_period_str(dp_id)

            for col in cv.get_columns():
                # skip zero-volume column
                if not col.get_volume():
                    continue

                i += 1
                node_seq = path_sep.join(
                    nodes[x].get_node_id() for x in reversed(col.nodes)
                )
                link_seq = path_sep.join(
                    links[x].get_link_id() for x in reversed(col.links)
                )

                geometry = ''
                if output_geometry:
                    geometry = ', '.join(
                        nodes[x].get_coordinate() for x in reversed(col.nodes)
                    )
                    geometry = 'LINESTRING (' + geometry + ')'

                line = [i,
                        oz_id,
                        dz_id,
                        col.get_id(),
                        at_str,
                        dp_str,
                        '{:.4f}'.format(col.get_volume()),
                        col.get_toll(),
                        col.get_travel_time(),
                        col.get_distance(),
                        node_seq,
                        link_seq,
                        geometry]

                writer.writerow(line)

        if output_dir == '.':
            print(f'\ncheck route_assignment.csv in {os.getcwd()} for path finding results')
        else:
            print(
                f'\ncheck route_assignment.csv in {os.path.join(os.getcwd(), output_dir)}'
                ' for path finding results'
            )


def output_link_performance(ui, mode='ue', output_dir='.'):
    with open(output_dir+'/link_performance.csv', 'w',  newline='') as fp:
        writer = csv.writer(fp)

        line = ['link_id',
                'from_node_id',
                'to_node_id',
                'time_period',
                'volume',
                'travel_time',
                'speed',
                'VOC',
                'geometry']

        if mode.lower().startswith('odme'):
            line.append('obs_count')
            line.append('deviation')

        writer.writerow(line)

        base = ui._base_assignment
        links = base.get_links()

        if not mode.lower().startswith('odme'):
            # UE or UE + Simulation
            for link in links:
                # connector
                if not link.length:
                    continue

                for dp in base.get_demand_periods():
                    avg_tt = link.get_period_avg_travel_time(dp.get_id())
                    speed = link.get_length() / (max(EPSILON, avg_tt) / 60)

                    line = [link.get_link_id(),
                            link.get_from_node_id(),
                            link.get_to_node_id(),
                            dp.get_period(),
                            link.get_period_flow_vol(dp.get_id()),
                            avg_tt,
                            speed,
                            link.get_period_voc(dp.get_id()),
                            link.get_geometry()]

                    writer.writerow(line)
        else:
            # ODME
            for link in links:
                # connector
                if not link.length:
                    continue

                for dp in base.get_demand_periods():
                    avg_tt = link.get_period_avg_travel_time(dp.get_id())
                    speed = link.get_length() / (max(EPSILON, avg_tt) / 60)

                    obs_count = ''
                    dev = ''
                    if dp.get_id() == 0:
                        obs_count = link.obs
                        dev = link.est_dev

                    line = [link.get_link_id(),
                            link.get_from_node_id(),
                            link.get_to_node_id(),
                            dp.get_period(),
                            link.get_period_flow_vol(dp.get_id()),
                            avg_tt,
                            speed,
                            link.get_period_voc(dp.get_id()),
                            link.get_geometry(),
                            obs_count,
                            dev]

                    writer.writerow(line)

        if output_dir == '.':
            print(f'check link_performance.csv in {os.getcwd()} for link performance')
        else:
            print(
                f'check link_performance.csv in {os.path.join(os.getcwd(), output_dir)}'
                ' for link performance'
            )


def output_agent_paths(ui, output_geometry=True, output_dir='.'):
    """ output unique agent path

    use it with find_path_for_agents() (which has been DEPRECATED)
    """
    with open(output_dir+'/agent_paths.csv', 'w',  newline='') as f:
        writer = csv.writer(f)

        line = ['agent_id',
                'o_zone_id',
                'd_zone_id',
                'origin node',
                'destination node',
                'path_id',
                'agent_type',
                'demand_period',
                'OD volume',
                'distance',
                'node_sequence',
                'link_sequence',
                'geometry']

        writer.writerow(line)

        base = ui._base_assignment
        nodes = base.get_nodes()
        agents = base.get_agents()
        agents.sort(key=lambda agent: agent.get_orig_node_id())

        at_str = ''
        dp_str = ''
        for a in agents:
            if not a.get_node_path():
               continue

            at_str = base.get_agent_type_str(a.get_at_id())
            dp_str = base.get_demand_period_str(a.get_dp_id())
            break

        pre_dest_node_id = -1
        for a in agents:
            if not a.get_node_path():
               continue

            if a.get_dest_node_id() == pre_dest_node_id:
                continue

            pre_dest_node_id = a.get_dest_node_id()
            id = a.get_id()

            at = a.get_at_id()
            dp = a.get_dp_id()
            oz = a.get_orig_zone_id()
            dz = a.get_dest_zone_id()

            vol = base.column_pool[(at, dp, oz, dz)].get_od_volume()

            geometry = ''
            if output_geometry:
                geometry = ', '.join(
                    nodes[x].get_coordinate() for x in reversed(a.get_node_path())
                )
                geometry = 'LINESTRING (' + geometry + ')'

            line = [id,
                    oz,
                    dz,
                    a.get_orig_node_id(),
                    a.get_dest_node_id(),
                    0,
                    at_str,
                    dp_str,
                    vol,
                    a.get_path_cost(),
                    base.get_agent_node_path(id, True),
                    base.get_agent_link_path(id, True),
                    geometry]

            writer.writerow(line)

        if output_dir == '.':
            print(f'\ncheck agent_paths.csv in {os.getcwd()} for unique agent paths')
        else:
            print(
                f'\ncheck agent_paths.csv in {os.path.join(os.getcwd(), output_dir)}'
                ' for unique agent paths'
            )


def output_synthetic_zones(ui, output_dir='.'):
    with open(output_dir+'/syn_zone.csv', 'w',  newline='') as f:
        writer = csv.writer(f)

        line = ['zone_id',
                'bin_index',
                'activity_nodes',
                'x_coord',
                'y_coord',
                'geometry',
                'production',
                'attraction']

        writer.writerow(line)

        base = ui._base_assignment
        network = base.network
        zones = network.zones

        for k, v in zones.items():
            bi = v.get_bin_index()
            nodes = '; '.join(str(x) for x in v.get_activity_nodes())
            # [U, D, L, R] = v.get_boundaries()
            x, y = v.get_coordinate()
            prod = v.get_production()
            geo = v.get_geo()

            line = [k, bi, nodes, x, y, geo, prod, prod]
            writer.writerow(line)

        if output_dir == '.':
            print(f'check zone.csv in {os.getcwd()} for synthetic zones')
        else:
            print(
                f'check zone.csv in {os.path.join(os.getcwd(), output_dir)}'
                ' for synthetic zones'
            )


def output_synthetic_demand(ui, output_dir='.'):
    with open(output_dir+'/syn_demand.csv', 'w',  newline='') as f:
        writer = csv.writer(f)

        line = ['o_zone_id', 'd_zone_id', 'volume']
        writer.writerow(line)

        column_pool = ui.get_column_pool()
        for k, v in column_pool.items():
            # k = (at_id, dp_id, oz_id, dz_id)
            line = [k[2], k[3], v.get_od_volume()]
            writer.writerow(line)

        if output_dir == '.':
            print(f'check demand.csv in {os.getcwd()} for synthetic demand\n')
        else:
            print(
                f'check demand.csv in {os.path.join(os.getcwd(), output_dir)}'
                ' for synthetic demand\n'
            )


def output_agent_trajectory(ui, output_dir='.'):
    with open(output_dir+'/trajectory.csv', 'w',  newline='') as f:
        writer = csv.writer(f)

        line = ['agent_id',
                'o_zone_id',
                'd_zone_id',
                'departure_time_in_min',
                'arrival_time_in_min',
                'complete_trip',
                'travel_time_in_min',
                'PCE',
                'distance',
                'node_sequence',
                'geometry',
                'time_sequence',
                'time_sequence_hhmmss']

        writer.writerow(line)

        A = ui._base_assignment
        nodes = A.get_nodes()
        agents = A.get_agents()
        st = A.get_simu_start_time()

        pre_dt = -1
        pre_od = -1, -1
        for a in agents:
            if a.get_node_path() is None:
                continue

            # do not output agents of the same OD pair with the same departure time
            if a.get_dep_time() == pre_dt and a.get_od() == pre_od:
                continue

            pre_dt = a.get_dep_time()
            pre_od = a.get_od()

            at = A.cast_interval_to_minute_float(a.link_arr_interval[-1]) + st
            dt = A.cast_interval_to_minute_float(a.link_dep_interval[-1]) + st
            time_seq1 = [at, dt]
            time_seq2 = [_get_time_stamp(at), _get_time_stamp(dt)]

            num = len(a.link_arr_interval) - 2
            # if num < 0, the following loop will be skipped
            for i in range(num, -1, -1):
                k = a.link_dep_interval[i]
                if k < 0:
                    break

                dt_ = A.cast_interval_to_minute_float(k) + st
                time_seq1.append(dt_)
                time_seq2.append(_get_time_stamp(dt_))

            time_seq1_str = ';'.join(str(t) for t in time_seq1)
            time_seq2_str = ';'.join(time_seq2)

            complete_trip = 'c'
            if a.link_dep_interval[0] < 0:
                complete_trip = 'n'

            # arrival time to the last node or departure time from the last link
            at_ = time_seq1[-1]
            # the original implementation using arrival time to the last link
            # to calculate trip time does not make sense
            tt = at_ - a.get_dep_time()

            node_path_str = A.get_agent_node_path(a.get_id(), True)
            geometry = ', '.join(
                nodes[x].get_coordinate() for x in reversed(a.get_node_path())
            )
            geometry = 'LINESTRING (' + geometry + ')'

            line = [a.get_id(),
                    a.get_orig_zone_id(),
                    a.get_dest_zone_id(),
                    a.get_dep_time(),
                    at_,
                    complete_trip,
                    tt,
                    a.PCE_factor,
                    a.get_path_cost(),
                    node_path_str,
                    geometry,
                    time_seq1_str,
                    time_seq2_str]

            writer.writerow(line)

        if output_dir == '.':
            print(f'\ncheck trajectory.csv in {os.getcwd()} for trajectories')
        else:
            print(
                f'\ncheck trajectory.csv in {os.path.join(os.getcwd(), output_dir)}'
                ' for trajectories'
            )


def read_measurements(ui, input_dir='.'):
    """ load traffic observations specified in measurement.csv """
    with open(input_dir+'/measurement.csv') as fp:
        print('read measurement.csv')

        base = ui._base_assignment
        map_id_to_no = base.network.map_id_to_no
        zones = base.network.zones

        links = base.get_links()
        # a temporary lookup table to retrieve a link using its head and tail
        link_lookup = {
            (link.from_node_no, link.to_node_no) : link for link in links
        }

        reader = csv.DictReader(fp)
        record_no = 0
        for line in reader:
            # get measurement type, which could be link, production, and attraction
            meas_type = line['measurement_type']
            if not meas_type:
                continue

            try:
                count = _convert_str_to_float(line['count'])
            except KeyError:
                count = _convert_str_to_float(line['count1'])
            except InvalidRecord:
                continue

            try:
                ub = line['upper_bound_flag']
            except KeyError:
                ub = 'false'

            is_upper_bounded = False
            ub_lowercase = ub.lower()
            if ub_lowercase.startswith('true') or ub_lowercase.startswith('1'):
                is_upper_bounded = True
                # all the other strings will be taken as False

            if meas_type.startswith('link'):
                from_node_id = line['from_node_id']
                to_node_id = line['to_node_id']

                try:
                    from_node_no = map_id_to_no[from_node_id]
                except KeyError:
                    print(
                        f'Exception: Node ID {from_node_id}'
                        ' NOT in the network!!'
                    )
                    continue

                try:
                    to_node_no = map_id_to_no[to_node_id]
                except KeyError:
                    print(
                        f'Exception: Node ID {to_node_id}'
                        ' NOT in the network!!'
                    )
                    continue

                # need to retrieve the link using from node and to node
                link_key = (from_node_no, to_node_no)
                if link_key not in link_lookup:
                    continue

                link = link_lookup[link_key]
                link.obs = count
                link.is_obs_upper_bounded = is_upper_bounded
            elif meas_type.startswith('production'):
                try:
                    zone_id = line['o_zone_id']
                except KeyError:
                    continue

                if not zone_id or zone_id not in zones:
                    continue

                zone = zones[zone_id]
                zone.prod_obs = count
                zone.is_prod_obs_upper_bounded = is_upper_bounded
            elif meas_type.startswith('attraction'):
                try:
                    zone_id = line['d_zone_id']
                except KeyError:
                    continue

                if not zone_id or zone_id not in zones:
                    continue

                zone = zones[zone_id]
                zone.attr_obs = count
                zone.is_attr_obs_upper_bounded = is_upper_bounded
            else:
                continue

            record_no += 1

        print(f'the number of valid measurements is {record_no}\n')


def read_demand(ui, use_synthetic_data = False, save_synthetic_data=True, input_dir='.'):
    """ a dedicated API to read demand and zone information """
    A = ui._base_assignment

    if not use_synthetic_data:
        # set up capacity ratio of affected links from special event
        for dp in A.demand_periods:
            se = dp.special_event
            if se is None:
                continue

            # k is link id and v is capacity ratio
            for k, v in se.get_affected_links():
                A.set_capacity_ratio(dp.get_id(), k, v)

        print('load demand')
        print('Step 1: try to load the default demand file: demand.csv')
        demand_loaded = False
        for d in A.get_demands():
            try:
                at = A.get_agent_type_id(d.get_agent_type_str())
                dp = A.get_demand_period_id(d.get_period())
                _read_demand(input_dir,
                             d.get_file_name(),
                             at,
                             dp,
                             A.network.zones,
                             A.column_pool)

                if not demand_loaded:
                    demand_loaded = True
            except FileNotFoundError:
                continue

        if demand_loaded:
            return

        print('the default demand files are NOT found!\n')

    # try to load the synthetic demand
    filename = 'syn_demand.csv'
    if use_synthetic_data:
        print(f'attempt to load the synthetic data: {filename} and syn_zone.csv')
    else:
        print(f'Step 2: attempt to load the synthetic data: {filename} and syn_zone.csv')

    for d in A.get_demands():
        try:
            at = A.get_agent_type_id(d.get_agent_type_str())
            dp = A.get_demand_period_id(d.get_period())
            _read_zones(ui)
            _read_demand(
                input_dir, filename, at, dp, A.network.zones, A.column_pool
            )
            # early termination to load only one synthetic demand file
            return
        except FileNotFoundError:
            break

    print('the synthetic data is missing or incomplete!\n')
    if use_synthetic_data:
        print('start to synthesize zones and demand!')
    else:
        print('Step 3: start to synthesize zones and demand!')

    # synthesize zones and demand
    network_to_zones(ui)
    print('data synthesis is complete!')

    if save_synthetic_data:
        output_synthetic_zones(ui, input_dir)
        output_synthetic_demand(ui, input_dir)


import os
from datetime import timedelta
from threading import Thread


__all__ = ['download_sample_data_sets', 'download_sample_setting_file']


# for precheck on connectivity of each OD pair
# 0: isolated, has neither outgoing links nor incoming links
# 1: has at least one outgoing link
# 2: has at least one incoming link
# 3: has both outgoing and incoming links
_zone_degrees = {}


class InvalidRecord(Exception):
    """a custom exception for invalid input from parsing a csv file"""
    pass


def _update_orig_zone(oz_id):
    if oz_id not in _zone_degrees:
        _zone_degrees[oz_id] = 1
    elif _zone_degrees[oz_id] == 2:
        _zone_degrees[oz_id] = 3


def _update_dest_zone(dz_id):
    if dz_id not in _zone_degrees:
        _zone_degrees[dz_id] = 2
    elif _zone_degrees[dz_id] == 1:
        _zone_degrees[dz_id] = 3


def _are_od_connected(oz_id, dz_id):
    connected = True

    # at least one node in O must have outgoing links
    if oz_id not in _zone_degrees or _zone_degrees[oz_id] == 2:
        connected = False
        print(f'WARNING! {oz_id} has no outgoing links to route volume '
              f'between OD: {oz_id} --> {dz_id}')

    # at least one node in D must have incoming links
    if dz_id not in _zone_degrees or _zone_degrees[dz_id] == 1:
        if connected:
            connected = False
        print(f'WARNING! {dz_id} has no incoming links to route volume '
              f'between OD: {oz_id} --> {dz_id}')

    return connected


# a little bit ugly
def _convert_str_to_int(s):
    if not s:
        raise InvalidRecord

    try:
        return int(s)
    except ValueError:
        # if s is not numeric, a ValueError will be then caught
        try:
            return int(float(s))
        except ValueError:
            raise InvalidRecord
    except TypeError:
        raise InvalidRecord


def _convert_str_to_float(s):
    if not s:
        raise InvalidRecord

    try:
        return float(s)
    except (TypeError, ValueError):
        raise InvalidRecord


def _convert_boundaries(bs):
    """a helper function to facilitate read_zones()"""
    if not bs:
        raise InvalidRecord

    prefix = 'LINESTRING ('
    postfix = ')'

    try:
        b = bs.index(prefix) + len(prefix)
        e = bs.index(postfix)
    except ValueError:
        raise Exception(f'Invalid Zone Boundaries: {bs}')

    bs_ = bs[b:e]
    vs = [x for x in bs_.split(',')]

    # validation
    if len(vs) != 5:
        raise Exception(f'Invalid Zone Boundaries: {bs}')

    if vs[0] != vs[-1]:
        raise Exception(f'Invalid Zone Boundaries: {bs}')

    L, U = vs[0].split(' ')
    R, U_ = vs[1].split(' ')
    if U != U_:
        raise Exception(
            f'Invalid Zone Boundaries: inconsistent upper boundary {U}; {U_}'
        )

    R_, D = vs[2].split(' ')
    if R != R_:
        raise Exception(
            'Invalid Zone Boundaries: inconsistent right boundary {R}; {R_}'
        )

    L_, D_ = vs[3].split(' ')
    if L != L_:
        raise Exception(
            'Invalid Zone Boundaries: inconsistent left boundary {L}; {L_}'
        )

    if D != D_:
        raise Exception(
            'Invalid Zone Boundaries: inconsistent lower boundary {D}; {D_}'
        )

    U = _convert_str_to_float(U)
    D = _convert_str_to_float(D)
    L = _convert_str_to_float(L)
    R = _convert_str_to_float(R)

    return U, D, L, R


def _get_time_stamp(minute):
    """ covert minute into HH:MM:SS as string """
    s = minute * 60
    return str(timedelta(seconds=s))


def _download_url(url, filename, loc_dir):
    try:
        import requests
    except ImportError:
        print('please print requests to proceed downloading!!')

    try:
        r = requests.get(url)
        r.raise_for_status()
        with open(loc_dir+filename, 'wb') as f:
            f.write(r.content)
    except requests.HTTPError:
        print(f'file not existing: {url}')
    except requests.ConnectionError:
        raise Exception('check your connection!!!')
    except Exception as e:
        raise e


def download_sample_data_sets():
    """ download sample data sets from the Github repo

    the following data sets will be downloaded: ASU, Braess Paradox, Chicago Sketch,
    Lima Network, Sioux Falls, and Two Corridors.
    """
    url = 'https://raw.githubusercontent.com/jdlph/Path4GMNS/dev/data/'

    data_sets = [
        "ASU",
        "Braess_Paradox",
        "Chicago_Sketch",
        "Lima_Network",
        "Sioux_Falls",
        "Two_Corridor"
    ]

    files = [
        "node.csv",
        "link.csv",
        "demand.csv",
        "settings.csv",
        "settings.yml"
    ]

    print('downloading starts')

    # data folder under cdw
    loc_data_dir = 'data'
    if not os.path.isdir(loc_data_dir):
        os.mkdir(loc_data_dir)

    for ds in data_sets:
        web_dir = url + ds + '/'
        loc_sub_dir = os.path.join(loc_data_dir, ds) + '/'

        if not os.path.isdir(loc_sub_dir):
            os.mkdir(loc_sub_dir)

        # multi-threading
        threads = []
        for x in files:
            t = Thread(
                target=_download_url,
                args=(web_dir+x, x, loc_sub_dir)
            )
            t.start()
            threads.append(t)

        for t in threads:
            t.join()

    print('downloading completes')
    print(f'check {os.path.join(os.getcwd(), loc_data_dir)} for downloaded data sets')


def download_sample_setting_file():
    """ download the sample settings.yml from the Github repo """
    url = 'https://raw.githubusercontent.com/jdlph/Path4GMNS/master/data/Chicago_Sketch/settings.yml'
    filename = 'settings.yml'
    loc_dir = './'

    _download_url(url, filename, loc_dir)

    print('downloading completes')
    print(f'check {os.getcwd()} for downloaded settings.yml')


import os
import csv
import threading

from .classes import AccessNetwork
from .path import single_source_shortest_path
from .consts import MAX_LABEL_COST, MIN_TIME_BUDGET, \
                    BUDGET_TIME_INTVL, MAX_TIME_BUDGET


__all__ = ['evaluate_accessibility', 'evaluate_equity']


def _get_interval_id(t):
    """ return interval id in predefined time budget intervals

    [0, MIN_TIME_BUDGET],

    (MIN_TIME_BUDGET + (i-1)*BUDGET_TIME_INTVL, MIN_TIME_BUDGET + i*BUDGET_TIME_INTVL]
        where, i is integer and i >= 1
    """
    if t < MIN_TIME_BUDGET:
        return 0

    if ((t-MIN_TIME_BUDGET) % BUDGET_TIME_INTVL) == 0:
        return int((t-MIN_TIME_BUDGET) / BUDGET_TIME_INTVL)

    return int((t-MIN_TIME_BUDGET) / BUDGET_TIME_INTVL) + 1


def _update_min_travel_time(an, at, min_travel_times, time_dependent, demand_period_id):
    an.update_generalized_link_cost(at, time_dependent, demand_period_id)

    at_str = at.get_type_str()
    max_min = 0
    for c in an.get_centroids():
        node_id = c.get_node_id()
        zone_id = c.get_zone_id()
        single_source_shortest_path(an, node_id)
        for c_ in an.get_centroids():
            if c_ == c:
                continue

            node_no = c_.get_node_no()
            to_zone_id = c_.get_zone_id()
            min_tt = an.get_node_label_cost(node_no)
            # this function will dramatically slow down the whole process
            min_dist = an.get_sp_distance(node_no)
            min_travel_times[(zone_id, to_zone_id, at_str)] = min_tt, min_dist

            if min_tt < MAX_LABEL_COST and max_min < min_tt:
                max_min = min_tt

    return max_min


def _output_od_accessibility(min_travel_times, zones, mode, output_dir):
    """ output accessibility for each OD pair (i.e., travel time) """
    with open(output_dir+'/od_accessibility.csv', 'w',  newline='') as f:
        headers = ['o_zone_id', 'd_zone_id', 'accessibility', 'distance', 'geometry']

        writer = csv.writer(f)
        writer.writerow(headers)

        # for multimodal case, find the minimum travel time
        # under mode 'a' (i.e., auto)
        for k, v in min_travel_times.items():
            # k = (from_zone_id, to_zone_id, at_type_str)
            if k[2] != mode:
                continue

            # output accessibility
            # no exception handlings here as min_travel_times is constructed
            # directly using an.get_centroids()
            coord_oz = zones[k[0]].get_coordinate_str()
            coord_dz = zones[k[1]].get_coordinate_str()
            geo = 'LINESTRING ()'
            if coord_oz and coord_dz:
                geo = 'LINESTRING (' + coord_oz + ', ' + coord_dz + ')'

            tt = v[0]
            dis = v[1]
            if tt >= MAX_LABEL_COST:
                tt = 'N/A'
                dis = 'N/A'

            line = [k[0], k[1], tt, dis, geo]
            writer.writerow(line)

        if output_dir == '.':
            print(f'check od_accessibility.csv in {os.getcwd()} for OD accessibility')
        else:
            print(
                f'check od_accessibility.csv in {os.path.join(os.getcwd(), output_dir)}'
                ' for OD accessibility'
            )


def _output_zone_accessibility(min_travel_times, interval_num,
                               zones, ats, output_dir):
    """ output zone accessibility matrix for each agent type """

    with open(output_dir+'/zone_accessibility.csv', 'w',  newline='') as f:
        time_budgets = [
            'TT_'+str(MIN_TIME_BUDGET+BUDGET_TIME_INTVL*i) for i in range(interval_num)
        ]

        headers = ['zone_id', 'geometry', 'mode']
        headers.extend(time_budgets)

        writer = csv.writer(f)
        writer.writerow(headers)

        # calculate accessibility
        for oz, v in zones.items():
            if not oz:
                continue

            for at in ats:
                at_str = at.get_type_str()
                # number of accessible zones from oz for each agent type
                counts = [0] * interval_num
                for dz in zones:
                    if (oz, dz, at_str) not in min_travel_times:
                        continue

                    min_tt = min_travel_times[(oz, dz, at_str)][0]
                    if min_tt >= MAX_LABEL_COST:
                        continue

                    id = _get_interval_id(min_tt)
                    while id < interval_num:
                        counts[id] += 1
                        id += 1
                # output accessibility

                # output the zone coordinates rather than the boundaries for the
                # following two reasons:
                # 1. to be consistent with _output_od_accessibility()
                # 2. v.get_geo() is always empty as no boundary info is provided
                #    in node.csv
                geo = 'LINESTRING ()'
                coord = v.get_coordinate_str()
                if coord:
                    geo = 'LINESTRING (' + coord + ')'

                line = [oz, geo, at.get_type_str()]
                line.extend(counts)
                writer.writerow(line)

        if output_dir == '.':
            print(f'check zone_accessibility.csv in {os.getcwd()} for zone accessibility')
        else:
            print(
                f'check zone_accessibility.csv in {os.path.join(os.getcwd(), output_dir)}'
                ' for zone accessibility'
            )


def _output_equity(output_dir, time_budget, equity_metrics, equity_zones):
    with open(output_dir+'/equity_'+str(time_budget)+'min.csv', 'w',  newline='') as f:
        headers = ['bin_index', 'mode', 'zones',
                   'min_accessibility', 'zone_id',
                   'max_accessibility', 'zone_id',
                   'mean_accessibility']
        writer = csv.writer(f)
        writer.writerow(headers)

        for k, v in sorted(equity_metrics.items()):
            try:
                avg = round(v[4] / len(equity_zones[k]), 2)
                zones = ', '.join(str(x) for x in equity_zones[k])
                line = [k[0], k[1], zones, v[0], v[1], v[2], v[3], avg]
            except ZeroDivisionError:
                continue

            writer.writerow(line)

        if output_dir == '.':
            print(
                f'\ncheck equity_{time_budget} min.csv in {os.getcwd()} for equity evaluation')
        else:
            print(
                f'\ncheck equity_{time_budget} min.csv in {os.path.join(os.getcwd(), output_dir)}'
                ' for equity evaluation')


def evaluate_accessibility(ui,
                           single_mode=False,
                           mode='auto',
                           time_dependent=False,
                           demand_period_id=0,
                           output_dir='.'):
    """ perform accessibility evaluation for a target mode or more

    Parameters
    ----------
    ui
        network object generated by pg.read_network()

    single_mode
        True or False. Its default value is False. It will only affect the
        output to zone_accessibility.csv.

        If False, the accessibility evaluation will be conducted
        for all the modes defined in settings.yml. The number of accessible
        zones from each zone under each defined mode given a budget time (up
        to 240 minutes) will be outputted to zone_accessibility.csv.

        If True, the accessibility evaluation will be only conducted against the
        target mode. The number of accessible zones from each zone under the
        target mode given a budget time (up to 240 minutes) will be outputted
        to zone_accessibility.csv.

    mode
        target mode with its default value as 'auto'. It can be
        either agent type or its name. For example, 'w' and 'walk' are
        equivalent inputs.

    time_dependent
        True or False. Its default value is False.

        If True, the accessibility will be evaluated using the period link
        free-flow travel time (i.e., VDF_fftt). In other words, the
        accessibility is time-dependent.

        If False, the accessibility will be evaluated using the link length and
        the free flow travel speed of each mode.

    demand_period_id
        The sequence number of demand period listed in demand_periods in
        settings.yml. demand_period_id of the first demand_period is 0.

        Use it with time_dependent when there are multiple demand periods. Its
        default value is 0.

    output_dir
        The directory path where zone_accessibility.csv and od_accessibility.csv
        are output. The default is the current working directory (CDW).

    Returns
    -------
    None

    Note
    ----
    The following files will be output.

    zone_accessibility.csv
        accessibility as the number of accessible zones from each
        zone for a target mode or any mode defined in settings.yml given a
        budget time (up to 240 minutes).

    od_accessibility.csv:
        accessibility between each OD pair in terms of free flow travel time.
    """
    base = ui._base_assignment
    an = AccessNetwork(base.network)
    ats = None

    zones = base.network.zones

    max_min = 0
    min_travel_times = {}
    at_name, at_str = base._convert_mode(mode)
    if not single_mode:
        ats = base.get_agent_types()
        for at in ats:
            an.set_target_mode(at.get_name())
            max_min_ = _update_min_travel_time(an,
                                               at,
                                               min_travel_times,
                                               time_dependent,
                                               demand_period_id)
            if max_min_ > max_min:
                max_min = max_min_
    else:
        an.set_target_mode(at_name)
        at = base.get_agent_type(at_str)
        max_min = _update_min_travel_time(an,
                                          at,
                                          min_travel_times,
                                          time_dependent,
                                          demand_period_id)
        ats = [at]

    interval_num = _get_interval_id(min(max_min, MAX_TIME_BUDGET)) + 1

    # multithreading to reduce output time
    t = threading.Thread(
        target=_output_od_accessibility,
        args=(min_travel_times, zones, at_str, output_dir)
    )
    t.start()

    t = threading.Thread(
        target=_output_zone_accessibility,
        args=(min_travel_times, interval_num, zones, ats, output_dir)
    )
    t.start()


def evaluate_equity(ui, single_mode=False, mode='auto', time_dependent=False,
                    demand_period_id=0, time_budget=60, output_dir='.'):
    """ evaluate equity for each zone under a time budget

    Parameters
    ----------
    ui
        network object generated by pg.read_network()

    single_mode
        True or False. Its default value is False. It will only affect the
        output to zone_accessibility.csv.

        If False, the equity evaluation will be conducted for all the modes defined
        in settings.yml.

        If True, the equity evaluation will be only conducted against the
        target mode.

    mode
        target mode with its default value as 'auto'. It can be
        either agent type or its name. For example, 'w' and 'walk' are
        equivalent inputs.

    time_dependent
        True or False. Its default value is False.

        If True, the accessibility will be evaluated using the period link
        free-flow travel time (i.e., VDF_fftt). In other words, the
        accessibility is time-dependent.

        If False, the accessibility will be evaluated using the link length and
        the free flow travel speed of each mode.

    demand_period_id
        The sequence number of demand period listed in demand_periods in
        settings.yml. demand_period_id of the first demand_period is 0.

        Use it with time_dependent when there are multiple demand periods. Its
        default value is 0.

    time_budget
        the amount of time to travel in minutes

    output_dir
        The directory path where the evaluation result is output. The default
        is the current working directory (CDW).

    Returns
    -------
    None

    Note
    ----
    The following file will be output.

    equity_str.csv
        equity statistics including minimum accessibility (and the corresponding
        zone), maximum accessibility (and the corresponding zone), and mean
        accessibility for each bin_index. The accessible zones will be output
        as well.

        str in the file name refers to the time budget. For example, the file
        name will be equity_60min.csv if the time budget is 60 min.
    """
    base = ui._base_assignment
    an = AccessNetwork(base.network)
    zones = an.base.zones
    ats = None

    min_travel_times = {}
    equity_metrics = {}
    equity_zones = {}

    if not single_mode:
        ats = base.get_agent_types()
        for at in ats:
            an.set_target_mode(at.get_name())
            _update_min_travel_time(an,
                                    at,
                                    min_travel_times,
                                    time_dependent,
                                    demand_period_id)
    else:
        at_name, at_str = base._convert_mode(mode)
        an.set_target_mode(at_name)
        at = base.get_agent_type(at_str)
        _update_min_travel_time(an,
                                at,
                                min_travel_times,
                                time_dependent,
                                demand_period_id)
        ats = [at]

    # v is zone object
    for oz, v in zones.items():
        if not oz:
            continue

        bin_index = v.get_bin_index()
        for at in ats:
            at_str = at.get_type_str()

            count = 0
            for dz in zones:
                if (oz, dz, at_str) not in min_travel_times:
                    continue

                min_tt = min_travel_times[(oz, dz, at_str)][0]
                if min_tt > time_budget:
                    continue

                count += 1

            if (bin_index, at_str) not in equity_metrics:
                equity_metrics[(bin_index, at_str)] = [count, oz, count, oz, 0]
                equity_zones[(bin_index, at_str)] = []
            equity_zones[(bin_index, at_str)].append(oz)

            # 0: min_accessibility, 1: zone_id, 2: max_accessibility,
            # 3: zone_id, 4: cumulative count,
            # where 0 to 4 are indices of each element of equity_metrics.
            if count < equity_metrics[(bin_index, at_str)][0]:
                equity_metrics[(bin_index, at_str)][0] = count
                equity_metrics[(bin_index, at_str)][1] = oz
            elif count > equity_metrics[(bin_index, at_str)][2]:
                equity_metrics[(bin_index, at_str)][2] = count
                equity_metrics[(bin_index, at_str)][3] = oz

            equity_metrics[(bin_index, at_str)][4] += count

    _output_equity(output_dir, time_budget, equity_metrics, equity_zones)